"""Upgrade from older versions"""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from tauon.t_modules.t_extra import RadioPlaylist, RadioStation, StarRecord, TauonPlaylist, TauonQueueItem, atomic_save

if TYPE_CHECKING:
	from pathlib import Path

	from tauon.t_modules.t_main import GuiVar, Prefs, Tauon, TrackClass


def migrate_star_store_71(tauon: Tauon) -> None:
	import pickle  # noqa: PLC0415

	backup_star_db = tauon.user_directory / "star.p.bak71"
	if not backup_star_db.exists():
		logging.info("Creating a backup Star database star.p.bak71")
		with atomic_save(backup_star_db) as file:
			pickle.dump(tauon.star_store.db, file, protocol=pickle.HIGHEST_PROTOCOL)

	new_starstore_db: dict[tuple[str, str, str], StarRecord] = {}
	old_record: list[int | str] = []  # Here just for typing
	for key, old_record in tauon.star_store.db.items():
		if isinstance(old_record, StarRecord):
			logging.warning(f"Record {old_record} was already a StarRecord, skipping this migration over…")
			break
		new_record = StarRecord()
		new_record.playtime = old_record[0]
		new_record.loved = "L" in old_record[1]
		new_record.rating = old_record[2]
		# There was a bug where the fourth element was not set
		if len(old_record) == 4:
			new_record.loved_timestamp = old_record[3]
		new_starstore_db[key] = new_record
	else:
		tauon.star_store.db = new_starstore_db
		logging.info("Saving newly migrated StarStore db…")
		with atomic_save(tauon.user_directory / "star.p") as file:
			pickle.dump(tauon.star_store.db, file, protocol=pickle.HIGHEST_PROTOCOL)


def database_migrate(
	*,
	tauon: Tauon,
	db_version: float,
	master_library: dict[int, TrackClass],
	legacy_track_misc: dict[int, dict],
	install_mode: bool,
	multi_playlist: list[str | int | bool] | list[TauonPlaylist],
	a_cache_dir: str,
	cache_directory: Path,
	config_directory: Path,
	install_directory: Path,
	user_directory: Path,
	gui: GuiVar,
	gen_codes: dict[int, str],
	prefs: Prefs,
	radio_playlists: list[dict[str, int | str | list[dict[str, str]]]] | list[RadioPlaylist],
	p_force_queue: list[int | bool] | list[TauonQueueItem],
	theme: int,
) -> tuple[
	dict[int, TrackClass],
	list[TauonPlaylist],
	list[TauonQueueItem],
	int,
	Prefs,
	GuiVar,
	dict[int, str],
	list[RadioPlaylist],
]:
	"""Migrate database to a newer version if we're behind

	Returns all the objects that could've been possibly changed:
		master_library, multi_playlist, p_force_queue, theme, prefs, gui, gen_codes, radio_playlists
	"""
	if db_version <= 0:
		logging.error("Called database_migrate with db_version equal to or below 0!")
		raise ValueError

	logging.warning(f"Running migrations as DB version was {db_version}!")

	if db_version <= 64:  # noqa: PLR2004
		logging.info("Updating database to version 65")

		if install_directory != config_directory and (config_directory / "input.txt").is_file():
			with (config_directory / "input.txt").open("a") as f:
				f.write("\nescape Escape\n")
				f.write("toggle-mute M Ctrl\n")

	if db_version <= 65:  # noqa: PLR2004
		logging.info("Updating database to version 66")

		if install_directory != config_directory and (config_directory / "input.txt").is_file():
			with (config_directory / "input.txt").open("a") as f:
				f.write("\ntoggle-artistinfo O Ctrl\n")
				f.write("cycle-theme ] Ctrl\n")
				f.write("cycle-theme-reverse [ Ctrl\n")

	if db_version <= 66:  # noqa: PLR2004
		logging.info("Updating database to version 67")
		for key, value in tauon.star_store.db.items():
			if len(value) == 3:
				value.append(0)
				tauon.star_store.db[key] = value

	if db_version <= 67:  # noqa: PLR2004
		logging.info("Updating database to version 68")
		for p in multi_playlist:
			if len(p) == 11:
				p.append(False)

	if db_version <= 68:  # noqa: PLR2004
		logging.info("Updating database to version 69")
		new_multi_playlist: list[TauonPlaylist] = []
		new_queue: list[TauonQueueItem] = []
		for playlist in multi_playlist:
			new_multi_playlist.append(
				TauonPlaylist(
					title=playlist[0],
					playing=playlist[1],
					playlist_ids=playlist[2],
					position=playlist[3],
					hide_title=playlist[4],
					selected=playlist[5],
					uuid_int=playlist[6],
					last_folder=playlist[7],
					hidden=playlist[8],
					locked=playlist[9],
					parent_playlist_id=playlist[10],
					persist_time_positioning=playlist[11],
				)
			)
		for queue in p_force_queue:
			new_queue.append(
				TauonQueueItem(
					track_id=queue[0],
					position=queue[1],
					playlist_id=queue[2],
					type=queue[3],
					album_stage=queue[4],
					uuid_int=queue[5],
					auto_stop=queue[6],
				)
			)
		multi_playlist = new_multi_playlist
		p_force_queue = new_queue

	if db_version <= 69:  # noqa: PLR2004
		logging.info("Updating database to version 70")
		new_radio_playlists: list[RadioPlaylist] = []
		for playlist in radio_playlists:
			stations: list[RadioStation] = []

			for station in playlist["items"]:
				stations.append(
					RadioStation(
						title=station["title"],
						stream_url=station["stream_url"],
						country=station.get("country", ""),
						website_url=station.get("website_url", ""),
						icon=station.get("icon", ""),
						stream_url_fallback=station.get("stream_url_unresolved", ""),
					)
				)
			new_radio_playlists.append(
				RadioPlaylist(
					uid=playlist["uid"], name=playlist["name"], scroll=playlist.get("scroll", 0), stations=stations
				)
			)
		radio_playlists = new_radio_playlists

	if db_version <= 71:  # This migration used both 71 and 72  # noqa: PLR2004
		logging.info("Updating database to version 72")
		migrate_star_store_71(tauon)

	if db_version <= 72:  # noqa: PLR2004
		# prefs.playlist_exports = save[168]
		logging.info("Updating database to version 73")
		for key, item in prefs.playlist_exports.items():
			playlist = None
			for p in multi_playlist:
				if p.uuid_int == key:
					playlist = p
					break
			else:
				continue

			if item.get("auto"):
				playlist.auto_export = True

			path = item.get("path")
			if path:
				if not path.endswith("/") and not path.endswith("\\"):
					path = path + "/"
				playlist.playlist_file = path

			type = item.get("type")
			if type:
				playlist.export_type = type

			relative = item.get("relative")
			if relative:
				playlist.relative_export = relative

	if db_version <= 73:  # noqa: PLR2004
		logging.info("Updating database to version 74")
		for playlist in multi_playlist:
			if not isinstance(playlist, TauonPlaylist):
				continue

			last_folder = playlist.last_folder
			if isinstance(last_folder, str):
				playlist.last_folder = [last_folder] if last_folder else []
			elif last_folder is None:
				playlist.last_folder = []
			elif isinstance(last_folder, list):
				playlist.last_folder = [str(path) for path in last_folder if path]
			else:
				try:
					playlist.last_folder = [str(path) for path in last_folder if path]
				except TypeError:
					playlist.last_folder = [str(last_folder)] if last_folder else []

	if db_version <= 74:  # noqa: PLR2004
		logging.info("Updating database to version 75")

	if db_version <= 75:  # noqa: PLR2004
		logging.info("Updating database to version 76")
		logging.info("Clearing search caches")
		tauon.search_string_cache.clear()
		tauon.search_dia_string_cache.clear()

	if db_version <= 76:  # noqa: PLR2004
		logging.info("Updating database to version 77")
		# state.p used to pickle the loaded Milkdrop preset as a pathlib.Path,
		# which made the save file OS-specific. It is stored as a plain string
		# now; the loader normalises either form, so this bump only forces the
		# post-migration save to rewrite state.p in the new format.

	if db_version <= 77:  # noqa: PLR2004
		logging.info("Updating database to version 78")
		# The default MB4/MB5 bindings changed from toggle-gallery /
		# toggle-right-panel to layout cycling. Edit just those lines in the
		# user's input.txt: a button the user rebound themselves stays as-is.
		keymap_file = config_directory / "input.txt"
		if install_directory != config_directory and keymap_file.is_file():
			try:
				lines = keymap_file.read_text(encoding="utf_8").splitlines()
				tokens = [line.split() for line in lines]
				new_binds = []
				if ["toggle-right-panel", "MB5"] in tokens and ["cycle-layouts", "MB5"] not in tokens:
					lines[tokens.index(["toggle-right-panel", "MB5"])] = "toggle-right-panel"
					new_binds.append("cycle-layouts MB5")
				if ["toggle-gallery", "MB4"] in tokens and ["cycle-layouts-reverse", "MB4"] not in tokens:
					lines[tokens.index(["toggle-gallery", "MB4"])] = "toggle-gallery"
					new_binds.append("cycle-layouts-reverse MB4")
				if new_binds:
					keymap_file.write_text("\n".join([*lines, *new_binds]) + "\n", encoding="utf_8")
					# The keymap loader thread starts before migrations run and
					# has likely already read the old file — reload so the new
					# bindings apply this launch, not just the next one.
					gui.keymaps.maps.clear()
					gui.keymaps.load()
			except Exception:
				logging.exception("Failed to update MB4/MB5 bindings in input.txt")

	if db_version <= 78:  # noqa: PLR2004
		logging.info("Updating database to version 79")
		# Extended track metadata used to live in a catch-all TrackClass.misc
		# dict; it now has dedicated __slots__ fields. Spread any misc data
		# preserved from the old save format into those fields.
		if legacy_track_misc:
			from tauon.t_modules.t_main import _MISC_TO_FIELD  # noqa: PLC0415
			for index, misc in legacy_track_misc.items():
				track = master_library.get(index)
				if track is None:
					continue
				for mk, mv in misc.items():
					field = _MISC_TO_FIELD.get(mk)
					if field is not None:
						setattr(track, field, mv)

	return master_library, multi_playlist, p_force_queue, theme, prefs, gui, gen_codes, radio_playlists



# Tauon Music Box - secondary window scaffold
#
# Provides on-screen windows that render with their own SDL renderer, *in
# addition to* the main window, sharing the same draw stack.
#
# This is the shared foundation for any UI that needs to live in its own OS
# window rather than being clipped to the main window:
#
#   * Context menus that can extend past the edge of the main window. This
#     matters most on Wayland, where a client cannot position a top-level
#     window at arbitrary screen coordinates - menus must instead be popup
#     surfaces (xdg_popup) anchored to the parent, which the compositor keeps
#     on-screen by flipping/sliding near display edges.
#   * In future, mini-mode views drawn alongside the main window instead of
#     replacing it.
#
# Each SecondaryWindow owns:
#   * an SDL window (a popup/menu child of the main window, so the compositor
#     handles on-screen constraint for us, or a borderless top-level)
#   * its own SDL renderer
#   * its own TDraw instance, and therefore its own glyph/text cache
#
# Drawing code renders into a SecondaryWindow exactly as it renders into the
# main window, by using the window's `ddt`. Image assets are shared across
# renderers via their per-renderer lazy texture cache (see LoadImageAsset /
# WhiteModImageAsset in t_main).

from __future__ import annotations

import ctypes
import logging
from typing import TYPE_CHECKING

import sdl3

from tauon.t_modules.t_draw import TDraw
from tauon.t_modules.t_extra import ColourRGBA

if TYPE_CHECKING:
	from tauon.t_modules.t_main import Tauon


def renderer_key(renderer: sdl3.LP_SDL_Renderer) -> int:
	"""Stable integer identity for an SDL renderer pointer.

	Used as a dict key so shared resources (image-asset textures) can be cached
	per renderer. ctypes pointer objects are not reliably hashable/comparable by
	address, so we cast to the raw pointer value.
	"""
	return ctypes.cast(renderer, ctypes.c_void_p).value or 0


class SecondaryWindow:
	"""A borderless popup window with its own renderer and draw context."""

	def __init__(self, tauon: Tauon, *, popup: bool = True, transparent: bool = True, high_dpi: bool = True, focusable: bool = True) -> None:
		self.tauon = tauon
		self.parent: sdl3.LP_SDL_Window = tauon.t_window
		self.popup = popup
		self.transparent = transparent
		self.high_dpi = high_dpi
		# A non-focusable window never steals keyboard focus from the parent, so
		# the main window stays focused and keeps receiving keyboard input while
		# the popup is up. Mouse focus is independent and still follows the
		# pointer, so the popup still gets its own mouse events. This is what a
		# context menu wants.
		self.focusable = focusable

		self.window: sdl3.LP_SDL_Window | None = None
		self.renderer: sdl3.LP_SDL_Renderer | None = None
		self.ddt: TDraw | None = None

		# All public sizes/positions are in *content pixels* - the same space the
		# main window draws in (gui.scale baked in, matching tauon.window_size).
		# SDL window sizing/positioning and mouse-motion events are in logical
		# *points*, so we convert through the parent window's pixel/point ratio.
		# `scale` is pixels-per-point; the input router multiplies popup event
		# coordinates by it to recover content-pixel coordinates.
		self.scale = 1.0
		self.w = 0
		self.h = 0
		self.pos: tuple[int, int] = (0, 0)
		# Last pointer position seen inside this window, in its own content
		# pixels (window-local). The menu hit-tests against this; it is kept
		# entirely separate from the main window's mouse position.
		self.last_local: tuple[int, int] = (0, 0)
		self.visible = False

	# --- lifecycle -------------------------------------------------------

	def _refresh_scale(self) -> None:
		pt_w, pt_h = ctypes.c_int(0), ctypes.c_int(0)
		px_w, px_h = ctypes.c_int(0), ctypes.c_int(0)
		sdl3.SDL_GetWindowSize(self.parent, ctypes.byref(pt_w), ctypes.byref(pt_h))
		sdl3.SDL_GetWindowSizeInPixels(self.parent, ctypes.byref(px_w), ctypes.byref(px_h))
		if self.high_dpi and pt_w.value > 0:
			self.scale = px_w.value / pt_w.value
		else:
			self.scale = 1.0

	def _to_points(self, value: int) -> int:
		return int(round(value / self.scale)) if self.scale else int(value)

	def _create(self, w: int, h: int, offset_x: int, offset_y: int) -> bool:
		flags = 0
		if self.high_dpi:
			flags |= sdl3.SDL_WINDOW_HIGH_PIXEL_DENSITY
		if self.transparent:
			flags |= sdl3.SDL_WINDOW_TRANSPARENT
		if not self.focusable:
			flags |= sdl3.SDL_WINDOW_NOT_FOCUSABLE

		if self.popup:
			# POPUP_MENU maps to xdg_popup on Wayland: the position is an offset
			# relative to the parent window, and the compositor constrains it to
			# the display (flipping/sliding near edges) - which is exactly the
			# edge handling we want, against the real screen rather than the
			# main window bounds.
			flags |= sdl3.SDL_WINDOW_POPUP_MENU
			self.window = sdl3.SDL_CreatePopupWindow(
				self.parent, self._to_points(offset_x), self._to_points(offset_y),
				self._to_points(w), self._to_points(h), flags,
			)
		else:
			flags |= sdl3.SDL_WINDOW_BORDERLESS | sdl3.SDL_WINDOW_RESIZABLE
			self.window = sdl3.SDL_CreateWindow(b"Tauon", self._to_points(w), self._to_points(h), flags)

		if not self.window:
			logging.error(f"SecondaryWindow: failed to create window - {sdl3.SDL_GetError()}")
			return False

		self.renderer = sdl3.SDL_CreateRenderer(self.window, None)
		if not self.renderer:
			logging.error(f"SecondaryWindow: failed to create renderer - {sdl3.SDL_GetError()}")
			sdl3.SDL_DestroyWindow(self.window)
			self.window = None
			return False

		sdl3.SDL_SetRenderDrawBlendMode(self.renderer, sdl3.SDL_BLENDMODE_BLEND)

		self.ddt = TDraw(self.renderer)
		self.ddt.scale = self.tauon.ddt.scale
		# Share the main context's font registry. Font handles (e.g. 412) are
		# primed once on the main ddt via prime_font(); the spec strings are
		# renderer-independent, and these dicts are mutated in place on rescale,
		# so sharing the references keeps the popup's fonts in sync for free.
		self.ddt.f_dict = self.tauon.ddt.f_dict
		self.ddt.font_desc_cache = self.tauon.ddt.font_desc_cache

		self.w = int(w)
		self.h = int(h)
		return True

	def show(self, w: int, h: int, offset_x: int, offset_y: int) -> bool:
		"""Create-or-resize the window, position it relative to the parent, show it.

		`w`/`h`/`offset_x`/`offset_y` are in content pixels; offsets are relative
		to the parent window's top-left, matching Wayland popup anchor semantics.
		"""
		self._refresh_scale()

		if self.window is None:
			if not self._create(w, h, offset_x, offset_y):
				return False
			self.pos = (int(offset_x), int(offset_y))
		else:
			# Only touch geometry when it actually changes: Wayland restricts
			# moving/resizing popup surfaces after they are mapped.
			if (int(w), int(h)) != (self.w, self.h):
				sdl3.SDL_SetWindowSize(self.window, self._to_points(w), self._to_points(h))
				self.w = int(w)
				self.h = int(h)
			if (int(offset_x), int(offset_y)) != self.pos:
				sdl3.SDL_SetWindowPosition(self.window, self._to_points(offset_x), self._to_points(offset_y))
				self.pos = (int(offset_x), int(offset_y))

		if self.ddt is not None:
			self.ddt.scale = self.tauon.ddt.scale

		# Only show on the visible transition: re-showing every frame re-grabs
		# input focus and churns window state.
		if not self.visible:
			# Seed the pointer to the popup's top-left until real motion arrives.
			self.last_local = (0, 0)
			sdl3.SDL_ShowWindow(self.window)
			self.visible = True
		return True

	def _release_input_grab(self) -> None:
		"""Release any window mouse grab and the global mouse capture.

		SDL_CaptureMouse is a *global* capture (not tied to this window), which
		SDL acquires during window operations such as resizing. Destroying the
		window does NOT release it, so it must be released explicitly or the OS
		keeps routing all mouse events to this window and the main window goes
		unresponsive.
		"""
		if self.window is not None:
			sdl3.SDL_SetWindowMouseGrab(self.window, False)
		sdl3.SDL_CaptureMouse(False)

	def hide(self) -> None:
		if self.window is not None and self.visible:
			self._release_input_grab()
			sdl3.SDL_HideWindow(self.window)
			# Hiding does not reliably hand input focus back to the parent, so
			# raise it explicitly.
			sdl3.SDL_RaiseWindow(self.parent)
		self.visible = False

	def destroy(self) -> None:
		self._release_input_grab()
		if self.renderer is not None:
			sdl3.SDL_DestroyRenderer(self.renderer)
			self.renderer = None
		if self.window is not None:
			sdl3.SDL_DestroyWindow(self.window)
			self.window = None
		self.ddt = None
		self.visible = False

	# --- per-frame drawing ----------------------------------------------

	def begin_frame(self, clear: ColourRGBA = ColourRGBA(0, 0, 0, 0)) -> None:
		"""Make this window's renderer current and clear it."""
		sdl3.SDL_SetRenderTarget(self.renderer, None)
		sdl3.SDL_SetRenderDrawColor(self.renderer, clear.r, clear.g, clear.b, clear.a)
		sdl3.SDL_RenderClear(self.renderer)
		if self.ddt is not None:
			self.ddt.new_frame()

	def end_frame(self) -> None:
		sdl3.SDL_RenderPresent(self.renderer)

	def window_id(self) -> int:
		if self.window is None:
			return 0
		return sdl3.SDL_GetWindowID(self.window)



"""Tauon Music Box - Basic Drawing and Text Drawing Functions Module"""

# Copyright © 2015-2019, Taiko2k captain(dot)gxj(at)gmail.com

#     This file is part of Tauon Music Box.
#
#     Tauon Music Box is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     Tauon Music Box is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Lesser General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with Tauon Music Box.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import annotations

import ctypes
import io
import logging
import math
from collections import OrderedDict
from ctypes import byref, c_bool, c_float, c_size_t
from typing import TYPE_CHECKING

import sdl3
from PIL import Image

from tauon.t_modules.t_extra import ColourRGBA, Timer, alpha_blend, coll_rect

if TYPE_CHECKING:
	from io import BytesIO

	from tauon.t_modules.t_main import Tauon

try:
	from jxlpy import JXLImagePlugin

	logging.info("Found jxlpy for JPEG XL support")
except ModuleNotFoundError:
	logging.warning("Unable to import jxlpy, JPEG XL support will be disabled.")
except Exception:
	logging.exception("Unknown error trying to import jxlpy, JPEG XL support will be disabled.")


import sys

if sys.platform == "win32":
	import os

	os.environ["PANGOCAIRO_BACKEND"] = "fc"

import cairo
import gi

gi.require_version("Pango", "1.0")
gi.require_version("PangoCairo", "1.0")
from gi.repository import Pango, PangoCairo


class QuickThumbnail:
	def __init__(self, tauon: Tauon) -> None:
		self.ddt = tauon.ddt
		self.renderer = tauon.renderer
		self.items: list[QuickThumbnail] = []
		self.queue: list[QuickThumbnail] = []
		self.rect = sdl3.SDL_FRect(0.0, 0.0)
		self.texture = None
		self.surface = None
		self.size = 50
		self.alive = False
		self.url = None

	def destruct(self) -> None:
		if self.surface:
			sdl3.SDL_DestroySurface(self.surface)
			self.surface = None
		if self.texture:
			sdl3.SDL_DestroyTexture(self.texture)
			self.texture = None
		self.alive = False

	def read_and_thumbnail(self, f: str, width: int, height: int) -> None:
		g = io.BytesIO()
		g.seek(0)
		im = Image.open(f)
		im.thumbnail((width, height), Image.Resampling.LANCZOS)
		im.save(g, "PNG")
		g.seek(0)
		self.surface = self.ddt.load_image(g)
		# self.items.append(self)
		self.alive = True

	def prime(self) -> None:
		texture = sdl3.SDL_CreateTextureFromSurface(self.renderer, self.surface)
		sdl3.SDL_DestroySurface(self.surface)
		self.surface = None
		tex_w = c_float(0.0)
		tex_h = c_float(0.0)
		sdl3.SDL_GetTextureSize(texture, byref(tex_w), byref(tex_h))
		self.rect.w = int(tex_w.value)
		self.rect.h = int(tex_h.value)
		self.texture = texture

	def draw(self, x: int, y: int) -> bool | None:
		if len(self.items) > 30:
			img = self.items[0]
			img.destruct()
			self.items.remove(img)
		if not self.alive:
			if self not in self.queue:
				self.queue.append(self)
			return False
		if not self.texture:
			self.prime()
		self.rect.x = round(x)
		self.rect.y = round(y)
		sdl3.SDL_RenderTexture(self.renderer, self.texture, None, self.rect)

		return True


perf = Timer()


class TDraw:
	def __init__(self, renderer: sdl3.LP_SDL_Renderer) -> None:
		# All
		self.renderer = renderer
		self.scale = 1
		self.force_subpixel_text = False

		# Drawing
		self.sdlrect = sdl3.SDL_FRect(10.0, 10.0, 10.0, 10.0)

		# Text and Fonts
		self.source_rect = sdl3.SDL_FRect(0.0, 0.0, 0.0, 0.0)
		self.dest_rect = sdl3.SDL_FRect(0.0, 0.0, 0.0, 0.0)

		self.surf = cairo.ImageSurface(cairo.FORMAT_ARGB32, 0, 0)
		self.context = cairo.Context(self.surf)
		# self.layout is used for measurement (get_text_wh); self.draw_layout is a
		# persistent layout for the render path, re-targeted per draw with
		# PangoCairo.update_layout instead of allocating a fresh layout on every
		# cache miss (avoids GObject alloc + GC churn while scrolling).
		self.layout = PangoCairo.create_layout(self.context)
		self.draw_layout = PangoCairo.create_layout(self.context)

		self.text_background_colour = ColourRGBA(0, 0, 0, 255)
		self.pretty_rect: tuple[int, int, int, int] | None = None
		self.real_bg: bool = False
		self.alpha_bg: bool = False
		self.force_gray: bool = False
		self.f_dict: dict[float, tuple[str, int, float]] = {}
		self.font_desc_cache: dict[float, Pango.FontDescription] = {}
		# Rendered-text textures, in LRU order (oldest first).
		self.ttc: OrderedDict[
			tuple[int, str, int, int, int, int, int, int, int, int],
			list[sdl3.SDL_FRect | sdl3.LP_SDL_Texture | int | bool],
		] = OrderedDict()
		self.text_texture_cache_bytes = 0
		# The item cap is recomputed every frame (see new_frame) to fit the text
		# actually on screen rather than a fixed guess. The main UI and the track
		# list are counted separately because the track list re-renders on its own
		# cadence (only when gui.pl_update fires); the cap is their sum plus 50%
		# headroom. max_text_texture_cache_bytes still bounds memory.
		self._frame_text_draws = 0      # non-track-list cacheable draws this frame
		self._tracklist_text_draws = 0  # cacheable draws in the last track-list render
		self._counting_tracklist = False
		self.max_text_texture_cache_items = 0
		self.max_text_texture_cache_bytes = 48 * 1024 * 1024
		self.max_text_texture_cache_item_bytes = 256 * 1024
		self.text_wh_cache: OrderedDict[tuple[str, int, int, bool], tuple[int, int]] = OrderedDict()
		self.max_text_wh_cache_items = 2048

		self.was_truncated = False

	def load_image(self, g: BytesIO) -> sdl3.LP_SDL_Surface:
		size = g.getbuffer().nbytes
		pointer = ctypes.c_void_p(ctypes.addressof(ctypes.c_char.from_buffer(g.getbuffer())))
		stream = sdl3.SDL_IOFromMem(pointer, c_size_t(size))
		return sdl3.IMG_Load_IO(stream, c_bool(True))

	def rect_s(self, rectangle: tuple[int, int, int, int], colour: ColourRGBA, thickness: int) -> None:
		sdl3.SDL_SetRenderDrawColor(self.renderer, colour.r, colour.g, colour.b, colour.a)
		x, y, w, h = (round(x) for x in rectangle)
		th = math.floor(thickness)
		self.sdlrect.x = x - th
		self.sdlrect.y = y - th
		self.sdlrect.w = th
		self.sdlrect.h = h + th
		sdl3.SDL_RenderFillRect(self.renderer, self.sdlrect)  # left
		self.sdlrect.x = x - th
		self.sdlrect.y = y + h
		self.sdlrect.w = w + th
		self.sdlrect.h = th
		sdl3.SDL_RenderFillRect(self.renderer, self.sdlrect)  # bottom
		self.sdlrect.x = x
		self.sdlrect.y = y - th
		self.sdlrect.w = w + th
		self.sdlrect.h = th
		sdl3.SDL_RenderFillRect(self.renderer, self.sdlrect)  # top
		self.sdlrect.x = x + w
		self.sdlrect.y = y
		self.sdlrect.w = th
		self.sdlrect.h = h + th
		sdl3.SDL_RenderFillRect(self.renderer, self.sdlrect)  # right

	def rect_si(self, rectangle: tuple[int, int, int, int], colour: ColourRGBA, thickness: int) -> None:
		sdl3.SDL_SetRenderDrawColor(self.renderer, colour.r, colour.g, colour.b, colour.a)
		x, y, w, h = (round(x) for x in rectangle)
		th = math.floor(thickness)
		self.sdlrect.x = x
		self.sdlrect.y = y
		self.sdlrect.w = th
		self.sdlrect.h = h
		sdl3.SDL_RenderFillRect(self.renderer, self.sdlrect)  # left
		self.sdlrect.x = x
		self.sdlrect.y = y + (h - th)
		self.sdlrect.w = w
		self.sdlrect.h = th
		sdl3.SDL_RenderFillRect(self.renderer, self.sdlrect)  # bottom
		self.sdlrect.x = x
		self.sdlrect.y = y
		self.sdlrect.w = w
		self.sdlrect.h = th
		sdl3.SDL_RenderFillRect(self.renderer, self.sdlrect)  # top
		self.sdlrect.x = x + (w - th)
		self.sdlrect.y = y
		self.sdlrect.w = th
		self.sdlrect.h = h
		sdl3.SDL_RenderFillRect(self.renderer, self.sdlrect)  # right

	def rect_a(self, location_xy: list[int], size_wh: list[int], colour: ColourRGBA) -> None:
		self.rect((location_xy[0], location_xy[1], size_wh[0], size_wh[1]), colour)

	def clear_rect(self, rectangle: tuple[int, int, int, int]) -> None:
		sdl3.SDL_SetRenderDrawBlendMode(self.renderer, sdl3.SDL_BLENDMODE_NONE)
		sdl3.SDL_SetRenderDrawColor(self.renderer, 0, 0, 0, 0)

		self.sdlrect.x = float(rectangle[0])
		self.sdlrect.y = float(rectangle[1])
		self.sdlrect.w = float(rectangle[2])
		self.sdlrect.h = float(rectangle[3])

		sdl3.SDL_RenderFillRect(self.renderer, self.sdlrect)
		sdl3.SDL_SetRenderDrawBlendMode(self.renderer, sdl3.SDL_BLENDMODE_BLEND)

	def rect(self, rectangle: tuple[int, int, int, int], colour: ColourRGBA) -> None:
		sdl3.SDL_SetRenderDrawColor(self.renderer, colour.r, colour.g, colour.b, colour.a)

		self.sdlrect.x = float(rectangle[0])
		self.sdlrect.y = float(rectangle[1])
		self.sdlrect.w = float(rectangle[2])
		self.sdlrect.h = float(rectangle[3])

		# if fill:
		sdl3.SDL_RenderFillRect(self.renderer, self.sdlrect)
		# else:
		# 	sdl3.SDL_RenderDrawRect(self.renderer, self.sdlrect)

	def bordered_rect(
		self, rectangle: tuple[int, int, int, int], fill_colour: ColourRGBA, outer_colour: ColourRGBA, border_size: int
	) -> None:
		self.sdlrect.x = round(rectangle[0]) - border_size
		self.sdlrect.y = round(rectangle[1]) - border_size
		self.sdlrect.w = round(rectangle[2]) + border_size + border_size
		self.sdlrect.h = round(rectangle[3]) + border_size + border_size
		sdl3.SDL_SetRenderDrawColor(self.renderer, outer_colour.r, outer_colour.g, outer_colour.b, outer_colour.a)
		sdl3.SDL_RenderFillRect(self.renderer, self.sdlrect)
		self.sdlrect.x = round(rectangle[0])
		self.sdlrect.y = round(rectangle[1])
		self.sdlrect.w = round(rectangle[2])
		self.sdlrect.h = round(rectangle[3])
		sdl3.SDL_SetRenderDrawColor(self.renderer, fill_colour.r, fill_colour.g, fill_colour.b, fill_colour.a)
		sdl3.SDL_RenderFillRect(self.renderer, self.sdlrect)

	def line(self, x1: int, y1: int, x2: int, y2: int, colour: ColourRGBA) -> None:
		sdl3.SDL_SetRenderDrawColor(self.renderer, colour.r, colour.g, colour.b, colour.a)
		sdl3.SDL_RenderLine(self.renderer, round(x1), round(y1), round(x2), round(y2))

	def get_text_w(self, text: str, font: int, height: bool = False) -> int:
		x, y = self.get_text_wh(text, font, 3000)
		if height:
			return y
		return x

	def clear_text_cache(self) -> None:
		for so in self.ttc.values():
			sdl3.SDL_DestroyTexture(so[1])

		self.ttc.clear()
		self.text_texture_cache_bytes = 0
		self.text_wh_cache.clear()

	def new_frame(self) -> None:
		"""Resize the text-texture cache to fit the text on screen.

		Called once at the start of each rendered frame. The cap is the number
		of cacheable text draws in the main UI last frame plus the track list's
		last render, with 50% headroom, so everything visible stays cached
		instead of thrashing against a fixed limit. The byte ceiling still
		bounds memory."""
		self.max_text_texture_cache_items = int((self._frame_text_draws + self._tracklist_text_draws) * 1.5)
		self._frame_text_draws = 0
		self._counting_tracklist = False

	def begin_tracklist_count(self) -> None:
		"""Start counting track-list text draws separately (see new_frame)."""
		self._tracklist_text_draws = 0
		self._counting_tracklist = True

	def end_tracklist_count(self) -> None:
		self._counting_tracklist = False

	def prime_font(self, name: str, size: float, user_handle: float, offset: int = 0) -> None:
		self.f_dict[user_handle] = (name + " " + str(size * self.scale), offset, size * self.scale)
		self.font_desc_cache[user_handle] = Pango.FontDescription(self.f_dict[user_handle][0])
		self.text_wh_cache.clear()

	def _font_description(self, font: int) -> Pango.FontDescription:
		cached = self.font_desc_cache.get(font)
		if cached is not None:
			return cached

		font_description = Pango.FontDescription(self.f_dict[font][0])
		self.font_desc_cache[font] = font_description
		return font_description

	def get_text_wh(self, text: str, font: int, max_x: int, wrap: bool = False) -> tuple[int, int] | None:
		cache_key = None
		if len(text) <= 256:
			cache_key = (text, font, max_x, wrap)
			cached = self.text_wh_cache.get(cache_key)
			if cached is not None:
				self.text_wh_cache.move_to_end(cache_key)
				return cached

		self.layout.set_font_description(self._font_description(font))
		self.layout.set_ellipsize(Pango.EllipsizeMode.END)
		self.layout.set_width(max_x * 1000)
		if wrap:
			self.layout.set_height(20000 * 1000)
		else:
			self.layout.set_height(0)

		try:
			self.layout.set_text(text, -1)
		except Exception:
			logging.exception(f"Exception in get_text_wh for: {text}")
			self.layout.set_text(text.encode("utf-8", "replace").decode("utf-8"), -1)

		result = self.layout.get_pixel_size()
		if cache_key is not None:
			self.text_wh_cache[cache_key] = result
			if len(self.text_wh_cache) > self.max_text_wh_cache_items:
				self.text_wh_cache.popitem(last=False)
		return result

	def get_y_offset(self, text: str, font: int, max_x: int, wrap: bool = False) -> int:
		"""HACKY"""
		self.layout.set_font_description(self._font_description(font))
		self.layout.set_ellipsize(Pango.EllipsizeMode.END)
		self.layout.set_width(max_x * 1000)
		if wrap:
			self.layout.set_height(20000 * 1000)
		else:
			self.layout.set_height(0)

		try:
			self.layout.set_text(text, -1)
		except Exception:
			logging.exception(f"Exception in get_y_offset for: {text}")
			self.layout.set_text(text.encode("utf-8", "replace").decode("utf-8"), -1)

		y_off = self.layout.get_baseline() / 1000
		y_off = round(round(y_off) - 13 * self.scale)  # 13 for compat with way text position used to work

		return y_off

	def __render_text(self, key: dict, x: int, y: int, range_top: int, range_height: int, align: int) -> None:
		sd = key

		if sd[3]:
			self.was_truncated = True

		if align == 1:
			sd[0].x = round(x) - sd[0].w

		elif align == 2:
			sd[0].x -= int(sd[0].w / 2)

		if range_height is not None and range_height < sd[0].h:
			if range_top < 0:
				range_top = 0

			if range_top > sd[0].h - range_height:
				range_top = sd[0].h - range_height

			self.source_rect.y = round(range_top)
			self.source_rect.w = sd[0].w
			self.source_rect.h = round(range_height)

			self.dest_rect.x = sd[0].x
			self.dest_rect.y = sd[0].y
			self.dest_rect.w = sd[0].w
			self.dest_rect.h = round(range_height)

			# sdl3.SDL_RenderCopyEx(self.renderer, sd[1], self.source_rect, self.dest_rect, 0, None, 0)
			sdl3.SDL_RenderTexture(self.renderer, sd[1], self.source_rect, self.dest_rect)
			return

		sdl3.SDL_RenderTexture(self.renderer, sd[1], None, sd[0])

	def __draw_text_cairo(
		self,
		location: list[int],
		text: str,
		colour: ColourRGBA,
		font: int,
		max_x: int,
		bg: ColourRGBA,
		align: int = 0,
		max_y: int | None = None,
		wrap: bool = False,
		range_top: int = 0,
		range_height: int | None = None,
		real_bg: bool = False,
		key: tuple[int, str, int, int, int, int, int, int, int, int] | None = None,
	) -> int:
		# perf.set()
		force_cache = False
		if key:
			force_cache = True

		self.was_truncated = False

		max_x += 12  # Hack
		max_x = round(max_x)

		alpha_bg = self.alpha_bg
		force_gray = self.force_gray
		# real_bg = True

		if bg.a < 200:
			alpha_bg = True
			force_gray = True

		x = round(location[0])
		y = round(location[1])

		if self.pretty_rect:
			w, h = self.get_text_wh(text, font, max_x, wrap)
			quick_box = [x, y, w, h]

			if align == 1:
				quick_box[0] = x - quick_box[2]
			elif align == 2:
				quick_box[0] -= int(quick_box[2] / 2)

			if coll_rect(self.pretty_rect, quick_box):
				# self.rect_r(quick_box, [0, 0, 0, 100], True)
				# if self.real_bg:
				# real_bg = True
				alpha_bg = True
			else:
				alpha_bg = False

		if alpha_bg:
			bg = ColourRGBA(0, 0, 0, 0)

		if max_y is not None:
			max_y = round(max_y)

		if len(text) == 0:
			return 0

		if key is None:
			key = (max_x, text, font, colour.r, colour.g, colour.b, colour.a, bg.r, bg.g, bg.b)

		if not real_bg or force_cache:
			if self._counting_tracklist:
				self._tracklist_text_draws += 1
			else:
				self._frame_text_draws += 1
			sd = self.ttc.get(key)
			if sd:
				sd[0].x = round(x)
				sd[0].y = round(y) - sd[2]

				self.__render_text(sd, x, y, range_top, range_height, align)
				self.ttc.move_to_end(key)

				if wrap:
					return sd[0].h
				return sd[0].w

		if not self.pretty_rect:  # Would have already done this if True
			w, h = self.get_text_wh(text, font, max_x, wrap)

		if w < 1:
			return 0

		h += 4  # Compensate for characters that drop past the baseline, Pango doesn't seem to allow for this

		if wrap:
			w = max_x + 1

		if font not in self.f_dict:
			logging.info(f"Font not loaded: {font!s}")
			return 10

		format = sdl3.SDL_PIXELFORMAT_ARGB8888 if alpha_bg else sdl3.SDL_PIXELFORMAT_XRGB8888
		surface = None
		surface_locked = False
		pixel_data = None
		data = None
		surf = None
		context = None
		layout = None
		c = None
		y_off = 0

		try:
			if real_bg:
				box = sdl3.SDL_Rect(x, y - self.get_y_offset(text, font, max_x, wrap), w, h)

				if align == 1:
					box.x = x - box.w
				elif align == 2:
					box.x -= int(box.w / 2)

				ssurf = sdl3.SDL_RenderReadPixels(self.renderer, box)
				if not ssurf:
					logging.warning("SDL_RenderReadPixels failed while rendering text background")
					return 0

				if ssurf.contents.format != format:
					surface = sdl3.SDL_ConvertSurface(ssurf, format)
					sdl3.SDL_DestroySurface(ssurf)
					if not surface:
						logging.warning("SDL_ConvertSurface failed while rendering text background")
						return 0
				else:
					surface = ssurf
			else:
				surface = sdl3.SDL_CreateSurface(w, h, format)
				if not surface:
					logging.warning("SDL_CreateSurface failed while rendering text")
					return 0
				ctypes.memset(surface.contents.pixels, 0, surface.contents.pitch * h)

			if not sdl3.SDL_LockSurface(surface):
				logging.warning("SDL_LockSurface failed while rendering text")
				return 0
			surface_locked = True

			pixel_size = surface.contents.pitch * h
			pixel_data = (ctypes.c_ubyte * pixel_size).from_address(surface.contents.pixels)
			data = memoryview(pixel_data)

			if alpha_bg:
				surf = cairo.ImageSurface.create_for_data(data, cairo.FORMAT_ARGB32, w, h, surface.contents.pitch)
			else:
				surf = cairo.ImageSurface.create_for_data(data, cairo.FORMAT_RGB24, w, h, surface.contents.pitch)

			context = cairo.Context(surf)

			if force_gray:
				options = context.get_font_options()
				options.set_antialias(cairo.ANTIALIAS_GRAY)
				# options.set_hint_style(cairo.HINT_STYLE_NONE)
				context.set_font_options(options)
			elif self.force_subpixel_text:
				options = context.get_font_options()
				# options.set_antialias(cairo.ANTIALIAS_NONE)
				# options.set_antialias(cairo.ANTIALIAS_GRAY)
				options.set_antialias(cairo.ANTIALIAS_SUBPIXEL)
				context.set_font_options(options)

			layout = self.draw_layout
			PangoCairo.update_layout(context, layout)
			layout.set_auto_dir(False)

			if max_y is not None:
				layout.set_ellipsize(Pango.EllipsizeMode.END)
				layout.set_width(max_x * 1000)
				layout.set_height(max_y * 1000)
			else:
				layout.set_ellipsize(Pango.EllipsizeMode.END)
				layout.set_width(max_x * 1000)

				extra = 0
				if wrap:  # Compensate for height measurement being 1-2 lines too short. Pango bug?
					extra = round(400000 * self.scale)

				layout.set_height(h * 1000 + extra)

			if not wrap and max_y is None:
				layout.set_height(-1)

			# Attributes don't seem to be implemented in gi?
			# attrs = Pango.AttrList()
			# attrs.insert(Pango.Attribute(Pango.Underline.SINGLE))
			# layout.set_attributes(attrs)

			context.rectangle(0, 0, w, h)

			if not real_bg and not alpha_bg:
				context.set_source_rgb(bg.r / 255, bg.g / 255, bg.b / 255)
				# context.set_source_rgba(0, 0, 0, 0)
				context.fill()

			context.set_source_rgb(colour.r / 255, colour.g / 255, colour.b / 255)

			# desc = Pango.FontDescription(self.f_dict[font][0])
			# desc.set_family("Arial")

			layout.set_font_description(self._font_description(font))

			try:
				layout.set_text(text, -1)
			except Exception:
				logging.exception(f"Text error on text: {text}")
				layout.set_text(text.encode("utf-8", "replace").decode("utf-8"), -1)

			# logging.info(layout.get_direction(0))

			y_off = layout.get_baseline() / 1000
			y_off = round(round(y_off) - 13 * self.scale)  # 13 for compat with way text position used to work

			PangoCairo.show_layout(context, layout)

			self.was_truncated = layout.is_ellipsized()
			surf.flush()
			surf.finish()
			surf = None
			layout = None
			context = None
			data.release()
			data = None
			pixel_data = None
			sdl3.SDL_UnlockSurface(surface)
			surface_locked = False

			# Here the background colour is keyed out allowing lines to overlap slightly
			if not real_bg and not alpha_bg:
				format_details = sdl3.SDL_GetPixelFormatDetails(format)
				ke = sdl3.SDL_MapRGB(format_details, None, bg.r, bg.g, bg.b)
				sdl3.SDL_SetSurfaceColorKey(surface, True, ke)

			c = sdl3.SDL_CreateTextureFromSurface(self.renderer, surface)
			if not c:
				logging.warning("SDL_CreateTextureFromSurface failed while rendering text")
				return 0

			sdl3.SDL_DestroySurface(surface)
			surface = None

			if alpha_bg:
				blend_mode = sdl3.SDL_ComposeCustomBlendMode(
					sdl3.SDL_BLENDFACTOR_ONE,
					sdl3.SDL_BLENDFACTOR_ONE_MINUS_SRC_ALPHA,
					sdl3.SDL_BLENDOPERATION_ADD,
					sdl3.SDL_BLENDFACTOR_ONE,
					sdl3.SDL_BLENDFACTOR_ONE_MINUS_SRC_ALPHA,
					sdl3.SDL_BLENDOPERATION_ADD,
				)
				sdl3.SDL_SetTextureBlendMode(c, blend_mode)
		finally:
			layout = None
			context = None
			if surf is not None:
				surf.finish()
			if data is not None:
				data.release()
			pixel_data = None
			if surface_locked and surface is not None:
				sdl3.SDL_UnlockSurface(surface)
			if surface is not None:
				sdl3.SDL_DestroySurface(surface)

		dst = sdl3.SDL_FRect(round(x), round(y))
		dst.w = round(w)
		dst.h = round(h)
		dst.y = round(y) - y_off

		pack = [dst, c, y_off, self.was_truncated]
		cache_item_bytes = w * h * 4
		texture_cached = False

		try:
			self.__render_text(pack, x, y, range_top, range_height, align)

			# Don't cache if using real background data
			cache_texture = not real_bg and (
				force_cache or cache_item_bytes <= self.max_text_texture_cache_item_bytes
			)
			if cache_texture:
				pack.append(cache_item_bytes)
				self.ttc[key] = pack
				self.text_texture_cache_bytes += cache_item_bytes
				texture_cached = True
				while self.ttc and (
					len(self.ttc) > self.max_text_texture_cache_items
					or self.text_texture_cache_bytes > self.max_text_texture_cache_bytes
				):
					old_key, so = self.ttc.popitem(last=False)
					sdl3.SDL_DestroyTexture(so[1])
					if len(so) > 4:
						self.text_texture_cache_bytes -= so[4]
		finally:
			if c is not None and not texture_cached:
				sdl3.SDL_DestroyTexture(c)

		if wrap:
			return dst.h
		return dst.w

	def text(
		self,
		location: list[int],
		text: str,
		colour: ColourRGBA,
		font: int,
		max_w: int = 4000,
		bg: ColourRGBA | None = None,
		range_top: int = 0,
		range_height: int | None = None,
		real_bg: bool = False,
		key: tuple[int, str, int, int, int, int, int, int, int, int] | None = None,
	) -> int | None:
		# logging.info((text, font))

		if not text:
			return 0

		max_w = max(1, max_w)

		if bg is None:
			bg = self.text_background_colour

		if colour.a != 255:
			colour = alpha_blend(colour, bg)
		align = 0
		if len(location) > 2:
			if location[2] == 1:
				align = 1
			if location[2] == 2:
				align = 2
			if location[2] == 4:
				max_h = None
				if len(location) > 4:
					max_h = location[4]

				return self.__draw_text_cairo(
					location,
					text,
					colour,
					font,
					location[3],
					bg,
					max_y=max_h,
					wrap=True,
					range_top=range_top,
					range_height=range_height,
				)

		return self.__draw_text_cairo(location, text, colour, font, max_w, bg, align, real_bg=real_bg, key=key)



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

if TYPE_CHECKING:
	from collections.abc import Callable
	from io import TextIOWrapper
	from pathlib import Path

	from sdl3 import LP_SDL_Renderer, LP_SDL_Window

	from tauon.__main__ import LogHistoryHandler


@dataclass
class Holder:
	"""Class that holds variables for forwarding them from __main__.py to t_main.py"""

	# fmt:off
	t_window:               LP_SDL_Window
	renderer:               LP_SDL_Renderer
	logical_size:           list[int] # X Y res
	window_size:            list[int] # X Y res
	maximized:              bool
	scale:                  float
	window_opacity:         float
	draw_border:            bool
	transfer_args_and_exit: Callable[[]] # transfer_args_and_exit() - TODO(Martin): This should probably be moved to extra module
	old_window_position:    tuple[int, int] | None # X Y res
	install_directory:      Path
	user_directory:         Path
	pyinstaller_mode:       bool
	phone:                  bool
	window_default_size:    tuple[int, int] # X Y res
	window_title:           bytes # t_title.encode("utf-8")
	fs_mode:                bool
	t_title:                str # "Tauon"
	n_version:              str # "7.9.0"
	t_version:              str # "v" + n_version
	t_id:                   str # "tauonmb" | "com.github.taiko2k.tauonmb"
	t_agent:                str # "TauonMusicBox/7.9.0"
	dev_mode:               bool
	instance_lock:          TextIOWrapper | None
	log:                    LogHistoryHandler
	# fmt:on



"""Tauon Music Box - Web interface module"""

# Copyright © 2015-2019, Taiko2k captain(dot)gxj(at)gmail.com

#     This file is part of Tauon Music Box.
#
#     Tauon Music Box is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     Tauon Music Box is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Lesser General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with Tauon Music Box.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import annotations

import hashlib
import html
import io
import json
import logging
import os
import socket
import struct
import subprocess
import time
from http.server import BaseHTTPRequestHandler, HTTPServer
from socketserver import ThreadingMixIn
from typing import TYPE_CHECKING

from tauon.t_modules.t_enums import Backend, PlayingState, StopMode
from tauon.t_modules.t_extra import Timer

try:
	from zeroconf import ServiceInfo, Zeroconf
except ModuleNotFoundError:
	ServiceInfo = None
	Zeroconf = None
except Exception:
	logging.exception("Failed to import zeroconf, remote API discovery will be disabled.")
	ServiceInfo = None
	Zeroconf = None

if TYPE_CHECKING:
	from typing import Any

	from tauon.t_modules.t_main import AlbumArt, GuiVar, PlayerCtl, Prefs, Strings, Tauon, TrackClass


REMOTE_API_PORT = 7814
REMOTE_API_SERVICE_TYPE = "_tauon-remote._tcp.local."


def get_local_ipv4_address() -> str | None:
	s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
	s.settimeout(0)
	try:
		# Doesn't need to be reachable, only needs to select the active local interface.
		s.connect(("10.255.255.255", 1))
		return s.getsockname()[0]
	except Exception:
		logging.exception("Failed to determine local IPv4 address for remote API discovery.")
		return None
	finally:
		s.close()


class RemoteAPIZeroconfPublisher:
	def __init__(self, port: int = REMOTE_API_PORT) -> None:
		self.port = port
		self.zeroconf: Zeroconf | None = None
		self.service_info: ServiceInfo | None = None

	def start(self) -> None:
		if Zeroconf is None or ServiceInfo is None:
			logging.info("zeroconf unavailable, skipping remote API discovery registration.")
			return

		address = get_local_ipv4_address()
		if address is None:
			logging.warning("Remote API discovery registration skipped because no local IPv4 address was available.")
			return

		host_name = socket.gethostname().split(".")[0].strip() or "Tauon"
		server_name = f"{host_name}.local."
		service_name = f"Tauon Remote ({host_name}).{REMOTE_API_SERVICE_TYPE}"
		properties = {
			"app": "Tauon",
			"path": "/api1",
			"version": "1",
		}

		self.zeroconf = Zeroconf()
		self.service_info = ServiceInfo(
			REMOTE_API_SERVICE_TYPE,
			service_name,
			addresses=[socket.inet_aton(address)],
			port=self.port,
			properties=properties,
			server=server_name,
		)
		try:
			self.zeroconf.register_service(self.service_info)
			logging.info(f"Registered Tauon remote API discovery on {address}:{self.port}")
		except Exception:
			logging.exception("Failed to register Tauon remote API discovery service.")
			self.stop()

	def stop(self) -> None:
		if self.zeroconf is None:
			return
		if self.service_info is not None:
			try:
				self.zeroconf.unregister_service(self.service_info)
			except Exception:
				logging.exception("Failed to unregister Tauon remote API discovery service.")
		try:
			self.zeroconf.close()
		except Exception:
			logging.exception("Failed to close zeroconf after remote API discovery shutdown.")
		finally:
			self.zeroconf = None
			self.service_info = None


class ThreadedHTTPServer(ThreadingMixIn, HTTPServer):
	pass


class VorbisMonitor:
	def __init__(self, tauon: Tauon) -> None:
		self.tauon = tauon
		self.reset()
		self.enable: bool = True
		self.synced: bool = False
		self.buffer = io.BytesIO()
		self.tries: int = 0

	def reset(self, tries: int = 0) -> None:
		self.enable = True
		self.synced = False
		self.buffer = io.BytesIO()
		self.tries = tries

	def input(self, data: bytes) -> None:
		if not self.enable:
			return

		b = self.buffer
		b.seek(0, io.SEEK_END)
		b.write(data)

		# Check there's enough data to decode header
		b.seek(0, io.SEEK_END)
		l = b.tell()
		if l < 128:
			logging.info("Not enough data to parse vorbis")
			return

		# Get page length
		b.seek(0, io.SEEK_SET)
		ogg = b.read(4)

		if ogg != b"OggS":
			f = data.find(b"OggS")
			self.reset(self.tries)
			if f > -1:
				logging.info("Ogg stream synced")
				data = data[f:]
				b = self.buffer
				b.write(data)
			else:
				self.tries += 1
				if self.tries > 100:
					logging.info("Giving up looking for OGG pages")
					self.enable = False
				return

			# self.enable = False
			# return

		b.seek(0, io.SEEK_SET)
		header = struct.unpack("<4sBBqIIiB", b.read(27))
		segment_count = header[7]

		# Ensure we can read the segment table
		seg_bytes = b.read(segment_count)
		if len(seg_bytes) < segment_count:
			return  # wait for more data

		segs = seg_bytes

		length = 0
		for s in segs:
			length += s

		length += 27 + header[7]

		if l > length:
			h = b.read(7)
			# logging.info(h)
			if h in {b"\x03vorbis", b"OpusTag"}:
				if h == b"OpusTag":
					b.seek(1, io.SEEK_CUR)

				vendor_length = int.from_bytes(b.read(4), byteorder="little")
				vendor = int.from_bytes(b.read(vendor_length), byteorder="little")
				comment_list_length = int.from_bytes(b.read(4), byteorder="little")

				found_tags = {}
				for i in range(comment_list_length):
					comment_length = int.from_bytes(b.read(4), byteorder="little")
					comment = b.read(comment_length)

					key, value = comment.decode().split("=", 1)

					if key == "title":
						found_tags["title"] = value
					if key == "artist":
						found_tags["artist"] = value
					if key == "year":
						found_tags["year"] = value
					if key == "album":
						found_tags["album"] = value

				line = ""
				if "title" in found_tags:
					line += found_tags["title"]
					if "artist" in found_tags:
						line = found_tags["artist"] + " - " + line

				if self.tauon.radiobox.parse_vorbis_okay():
					self.tauon.pctl.found_tags = found_tags
					self.tauon.pctl.tag_meta = line

				logging.info("Found vorbis comment")
				logging.info(line)

			# Consume page from buffer
			b.seek(length, io.SEEK_SET)
			new = io.BytesIO()
			new.write(b.read())
			self.buffer = new


def send_file(path: str, mime: str, server) -> None:
	range_req = False
	start = 0
	end = None
	range_header = server.headers.get("Range")

	try:
		with open(path, "rb") as f:
			f.seek(0, 2)
			length = f.tell()

			if range_header:
				try:
					unit, value = range_header.split("=", 1)
					if unit.strip().lower() != "bytes" or "," in value:
						raise ValueError
					start_str, end_str = value.split("-", 1)
					if not start_str and not end_str:
						raise ValueError
					if start_str:
						start = int(start_str)
						if start < 0:
							raise ValueError
						end = int(end_str) if end_str else length - 1
						if end < 0:
							raise ValueError
					else:
						# Suffix ranges: bytes=-500 means last 500 bytes
						suffix_length = int(end_str)
						if suffix_length <= 0:
							raise ValueError
						start = max(0, length - suffix_length)
						end = length - 1
					range_req = True
				except (IndexError, ValueError):
					range_req = False
					start = 0
					end = None

			if range_req:
				if start >= length or end is None or end < start:
					server.send_response(416)
					server.send_header("Accept-Ranges", "bytes")
					server.send_header("Content-Range", f"bytes */{length}")
					server.end_headers()
					return
				if end >= length:
					end = length - 1
				chunk_length = end - start + 1
				server.send_response(206)
				server.send_header("Accept-Ranges", "bytes")
				server.send_header("Content-Range", f"bytes {start}-{end}/{length}")
				server.send_header("Content-Length", str(chunk_length))
				server.send_header("Content-Type", mime)
				f.seek(start)
			else:
				server.send_response(200)
				server.send_header("Accept-Ranges", "bytes")
				server.send_header("Content-Type", mime)
				server.send_header("Content-Length", str(length))
				f.seek(0)

			server.end_headers()

			remaining = end - start + 1 if range_req else None
			while True:
				read_size = 65536 if remaining is None else min(65536, remaining)
				if read_size <= 0:
					break
				data = f.read(read_size)
				if not data:
					break
				server.wfile.write(data)
				if remaining is not None:
					remaining -= len(data)
	except OSError:
		server.send_response(404)
		server.end_headers()


def webserve(
	pctl: PlayerCtl,
	prefs: Prefs,
	gui: GuiVar,
	album_art_gen: AlbumArt,
	install_directory: str,
	strings: Strings,
	tauon: Tauon,
) -> int | None:
	if prefs.enable_web is False:
		return 0

	gui.web_running = True

	class Server(BaseHTTPRequestHandler):
		def log_message(self, format: str, *args) -> None:
			logging.info(format % args)

		def send_file(self, path: str, mime: str) -> None:
			self.send_response(200)
			self.send_header("Content-type", mime)
			self.end_headers()

			with open(path, "rb") as f:
				self.wfile.write(f.read())

		def get_track_id(self, track: TrackClass) -> str:
			return hashlib.md5((str(track.index) + track.title + track.artist).encode()).hexdigest()

		def do_GET(self) -> None:
			path = self.path

			# logging.info(self.headers)
			# logging.info(path)

			if path == "/listenalong/":
				self.send_response(302)
				self.send_header("Location", "/listenalong")
				self.end_headers()

			elif path == "/listenalong":
				self.send_file(install_directory + "/templates/radio.html", "text/html")
			elif path == "/favicon.png":
				self.send_file(install_directory + "/assets/icon-64.png", "image/png")
			elif path == "/listenalong/play.svg":
				self.send_file(install_directory + "/templates/play.svg", "image/svg+xml")
			elif path == "/listenalong/pause.svg":
				self.send_file(install_directory + "/templates/pause.svg", "image/svg+xml")
			elif path == "/listenalong/stop.svg":
				self.send_file(install_directory + "/templates/stop.svg", "image/svg+xml")
			elif path == "/radio/radio.js":
				self.send_file(install_directory + "/templates/radio.js", "application/javascript")
			elif path == "/radio/theme.css":
				self.send_file(install_directory + "/templates/theme.css", "text/css")
			elif path == "/radio/logo-bg.png":
				self.send_file(install_directory + "/templates/logo-bg.png", "image/png")

			elif path.startswith("/llapi/audiofile/"):
				value = path[17:]
				track = pctl.playing_object()
				if not track:
					self.send_response(403)
					self.end_headers()
					return
				sid = self.get_track_id(track)
				if sid != value or track.is_network:
					self.send_response(403)
					self.end_headers()
					return
				if not track.fullpath or not os.path.isfile(track.fullpath):
					self.send_response(404)
					self.end_headers()
					return
				mime = "audio/mpeg"
				if track.file_ext == "FLAC":
					mime = "audio/flac"
				if track.file_ext in {"OGG", "OPUS", "OGA"}:
					mime = "audio/ogg"
				if track.file_ext == "M4A":
					mime = "audio/mp4"
				send_file(track.fullpath, mime, self)

			elif path == "/llapi/poll":
				self.send_response(200)
				self.send_header("Content-type", "application/json")

				ip = self.client_address[0]
				timer = tauon.listen_alongers.get(ip)
				if not timer:
					tauon.listen_alongers[ip] = Timer()
				else:
					timer.set()

				track = pctl.playing_object()
				if track is None:
					data = {"status": 0}
				else:
					data = {
						"status": pctl.playing_state,
						"id": self.get_track_id(track),
						"position": pctl.playing_time,
						"duration": track.length,
						"title": track.title,
						"artist": track.artist,
					}
				self.end_headers()
				data = json.dumps(data).encode()
				self.wfile.write(data)

			elif path.startswith("/llapi/picture/"):
				value = path[15:]
				track = pctl.playing_object()
				if not track:
					self.send_response(403)
					self.end_headers()
					return
				sid = self.get_track_id(track)
				if sid != value or track.is_network:
					self.send_response(403)
					self.end_headers()
					return

				try:
					base64 = album_art_gen.get_base64(track, (300, 300)).decode()
					data = {"image_data": base64}
				except Exception:
					logging.exception("No image data")
					data = {"image_data": "None"}

				lyrics = tauon.synced_to_static_lyrics.get(track)
				lyrics = html.escape(lyrics).replace("\r\n", "\n").replace("\r", "\n").replace("\n", "<br>")
				data["lyrics"] = lyrics

				data = json.dumps(data).encode()
				self.send_response(200)
				self.send_header("Content-type", "application/json")
				self.send_header("Content-length", str(len(data)))
				self.end_headers()
				self.wfile.write(data)

			else:
				self.send_response(404)
				self.end_headers()
				self.wfile.write(b"404 Not found")

	try:
		httpd = ThreadedHTTPServer(("0.0.0.0", prefs.metadata_page_port), Server)
		tauon.radio_server = httpd
		httpd.serve_forever()
		httpd.server_close()
	except OSError as e:
		if str(e) == "[Errno 98] Address already in use":
			logging.error("Not starting radio page server, is another Tauon instance already running?")  # noqa: TRY400
		else:
			logging.exception("Unknown OSError starting radio page server!")
	except Exception:
		logging.exception("Failed starting radio page server!")


def webserve2(pctl: PlayerCtl, album_art_gen: AlbumArt, tauon: Tauon) -> None:
	play_timer = Timer()
	discovery_publisher = RemoteAPIZeroconfPublisher()

	class Server(BaseHTTPRequestHandler):
		def log_message(self, format: str, *args) -> None:
			logging.info(format % args)

		def run_command(self, callback) -> None:
			self.send_response(200)
			# self.send_header("Content-type", "application/json")
			self.end_headers()
			callback()
			self.wfile.write(b"OK")

		def stream_opus_file(self, track: TrackClass) -> None:
			ffmpeg_path = tauon.get_ffmpeg()
			if ffmpeg_path is None:
				self.send_response(503)
				self.end_headers()
				self.wfile.write(b"ffmpeg unavailable")
				return

			command = [str(ffmpeg_path), "-v", "error"]
			if track.start_time:
				command.extend(["-ss", str(track.start_time)])
				if track.length > 0:
					command.extend(["-t", str(track.length)])
			command.extend([
				"-i", track.fullpath,
				"-vn",
				"-c:a", "libopus",
				"-b:a", "84k",
				"-f", "ogg",
				"-",
			])

			try:
				encoder = subprocess.Popen(
					command,
					stdin=subprocess.DEVNULL,
					stdout=subprocess.PIPE,
					stderr=subprocess.DEVNULL,
				)
			except OSError:
				logging.exception("Failed to start ffmpeg for /api1/fileopus")
				self.send_response(500)
				self.end_headers()
				self.wfile.write(b"Transcode start failed")
				return

			if encoder.stdout is None:
				self.send_response(500)
				self.end_headers()
				self.wfile.write(b"Transcode stream unavailable")
				return

			self.send_response(200)
			self.send_header("Content-type", "audio/ogg")
			self.send_header("Content-Disposition", 'attachment; filename="track.opus"')
			self.send_header("Connection", "close")
			self.end_headers()
			self.close_connection = True

			try:
				while True:
					data = encoder.stdout.read(65536)
					if not data:
						break
					self.wfile.write(data)
			except (BrokenPipeError, ConnectionResetError):
				pass
			finally:
				encoder.stdout.close()
				if encoder.poll() is None:
					encoder.terminate()
					try:
						encoder.wait(timeout=1)
					except subprocess.TimeoutExpired:
						encoder.kill()
						encoder.wait(timeout=1)

		def toggle_album_shuffle(self) -> None:
			pctl.album_shuffle_mode ^= True
			tauon.gui.update += 1

		def parse_trail(self, text: str) -> tuple[list[str], dict[str, str]]:
			params: dict[str, str] = {}
			both = text.split("?")
			levels = both[0].split("/")
			if len(both) > 1:
				pairs = both[1].split("&")
				for p in pairs:
					if "=" not in p:
						continue
					aa, bb = p.split("=", 1)
					params[aa] = bb

			return levels, params

		def get_track(
			self,
			track_position: int,
			playlist_index: int | None = None,
			track: TrackClass | None = None,
			album_id: int = -1,
		) -> dict[str, int | str | bool]:
			if track is None:
				if playlist_index is None:
					playlist = pctl.multi_playlist[pctl.active_playlist_playing].playlist_ids
				else:
					playlist = pctl.multi_playlist[playlist_index].playlist_ids
				track_id = playlist[track_position]
				track = pctl.get_track(track_id)

			data: dict[str, int | str | bool] = {}
			data["title"] = track.title
			data["artist"] = track.artist
			data["album"] = track.album
			data["album_artist"] = track.album_artist
			if not track.album_artist:
				data["album_artist"] = track.artist
			data["duration"] = int(track.length * 1000)
			data["id"] = track.index
			data["position"] = track_position
			data["path"] = track.fullpath
			data["album_id"] = album_id
			data["has_lyrics"] = track.lyrics != ""
			data["track_number"] = str(track.track_number).lstrip("0")
			data["can_download"] = not track.is_cue and not track.is_network

			return data

		def do_GET(self) -> None:
			path = self.path
			# logging.info(self.headers)
			if tauon.remote_limited and not tauon.chrome_mode:
				self.send_response(404)
				self.end_headers()
				self.wfile.write(b"404 Not found")
				return
			if tauon.remote_limited and (
				not path.startswith("/api1/pic/medium/")
				and not path.startswith("/api1/file/")
				and not path.startswith("/api1/fileopus")
			):
				self.send_response(404)
				self.end_headers()
				self.wfile.write(b"404 Not found")
				return

			if path.startswith("/api1/pic/small/"):
				value = path[16:]
				if value.isdigit() and int(value) in pctl.master_library:
					track = pctl.get_track(int(value))
					raw = album_art_gen.save_thumb(track, (75, 75), "")
					if raw:
						self.send_response(200)
						self.send_header("Content-type", "image/jpg")
						self.end_headers()
						self.wfile.write(raw.read())
					else:
						self.send_response(404)
						self.end_headers()
						self.wfile.write(b"No image found")

				else:
					self.send_response(404)
					self.end_headers()
					self.wfile.write(b"Invalid parameter")

			elif path.startswith("/api1/pic/medium/"):
				value = path[17:]
				logging.info(value)
				if value.isdigit() and int(value) in pctl.master_library:
					track = pctl.get_track(int(value))
					raw = album_art_gen.save_thumb(track, (1000, 1000), "")
					if raw:
						self.send_response(200)
						self.send_header("Content-type", "image/jpg")
						self.end_headers()
						self.wfile.write(raw.read())
					else:
						self.send_response(404)
						self.end_headers()
						self.wfile.write(b"No image found")

				else:
					self.send_response(404)
					self.end_headers()
					self.wfile.write(b"Invalid parameter")

			elif path.startswith("/api1/lyrics/"):
				value = path[13:]
				if value.isdigit() and int(value) in pctl.master_library:
					track = pctl.get_track(int(value))
					data = {}
					data["track_id"] = track.index
					data["lyrics_text"] = track.lyrics

					self.send_response(200)
					self.send_header("Content-type", "application/json")
					self.end_headers()
					data = json.dumps(data).encode()
					self.wfile.write(data)
				else:
					self.send_response(404)
					self.end_headers()
					self.wfile.write(b"Invalid parameter")

			# elif path.startswith("/api1/stream/"):
			# param = path[13:]
			#
			# if param.isdigit() and int(param) in pctl.master_library:
			# 	track = pctl.master_library[int(param)]
			# 	mime = "audio/mpeg"
			# 	#mime = "audio/ogg"
			# 	self.send_response(200)
			# 	self.send_header("Content-type", mime)
			# 	self.end_headers()
			#
			# 	cmd = ["ffmpeg", "-i", track.fullpath, "-c:a", "libopus", "-f", "ogg", "-"]
			# 	#cmd = ["ffmpeg", "-i", track.fullpath, "-c:a", "libvorbis", "-f", "ogg", "-"]
			# 	#cmd = ["ffmpeg", "-i", track.fullpath, "-c:a", "libmp3lame", "-f", "mp3", "-"]
			# 	encoder = subprocess.Popen(cmd, stdin=subprocess.PIPE, stdout=subprocess.PIPE)
			# 	while True:
			# 		data = encoder.stdout.read(1024)
			# 		if data:
			# 			self.wfile.write(data)
			elif path.startswith("/api1/playinghit/"):
				param = path[17:]
				if param.isdigit() and int(param) in pctl.master_library:
					t = play_timer.hit()
					if 0 < t < 5:
						tauon.star_store.add(int(param), t)

				self.send_response(200)
				self.send_header("Content-type", "text/plain")
				self.end_headers()
				self.wfile.write(b"OK")

			elif path.startswith("/api1/file/"):
				param = path[11:]

				# logging.info(self.headers)
				play_timer.hit()

				if param.isdigit() and int(param) in pctl.master_library:
					track = pctl.master_library[int(param)]
					if not track.fullpath or not os.path.isfile(track.fullpath):
						self.send_response(404)
						self.end_headers()
						self.wfile.write(b"File unavailable")
					else:
						mime = "audio/mpeg"
						if track.file_ext == "FLAC":
							mime = "audio/flac"
						if track.file_ext in {"OGG", "OPUS", "OGA"}:
							mime = "audio/ogg"
						if track.file_ext == "M4A":
							mime = "audio/mp4"
						send_file(track.fullpath, mime, self)
				else:
					self.send_response(404)
					self.end_headers()
					self.wfile.write(b"Invalid parameter")

			elif path.startswith("/api1/fileopus/"):
				param = path[15:].split("?", 1)[0]

				play_timer.hit()

				if param.isdigit() and int(param) in pctl.master_library:
					track = pctl.master_library[int(param)]
					if not track.fullpath or not os.path.isfile(track.fullpath):
						self.send_response(404)
						self.end_headers()
						self.wfile.write(b"File unavailable")
					else:
						self.stream_opus_file(track)
				else:
					self.send_response(404)
					self.end_headers()
					self.wfile.write(b"Invalid parameter")

			elif path.startswith("/api1/start/"):
				levels, _ = self.parse_trail(path)
				if len(levels) == 5:
					playlist = levels[3]
					position = levels[4]
					if playlist.isdigit() and position.isdigit():
						position = int(position)
						playlist = int(playlist)
						pl = tauon.id_to_pl(int(playlist))
						if pl is not None and pl < len(pctl.multi_playlist):
							playlist = pctl.multi_playlist[pl].playlist_ids
							if position < len(playlist):
								tauon.switch_playlist(pl, cycle=False, quiet=True)
								pctl.jump(playlist[position], position)

				self.send_response(200)
				self.send_header("Content-type", "text/plain")
				self.end_headers()
				self.wfile.write(b"OK")

			elif path == "/api1/play":
				self.run_command(tauon.pctl.play)
			elif path == "/api1/pause":
				self.run_command(tauon.pctl.pause_only)
			elif path == "/api1/stop":
				self.run_command(tauon.pctl.stop)
			elif path == "/api1/next":
				self.run_command(tauon.pctl.advance)
			elif path == "/api1/back":
				self.run_command(tauon.pctl.back)
			elif path == "/api1/shuffle":
				self.run_command(tauon.toggle_random)
			elif path == "/api1/album-shuffle":
				self.run_command(self.toggle_album_shuffle)
			elif path == "/api1/repeat":
				self.run_command(tauon.toggle_repeat)
			elif path == "/api1/auto-stop":
				if tauon.pctl.stop_mode == StopMode.OFF:
					tauon.pctl.stop_mode = StopMode.TRACK
				else:
					tauon.pctl.stop_mode = StopMode.OFF
				tauon.gui.update += 1
			elif path == "/api1/version":
				data = {"version": 1}
				self.send_response(200)
				self.send_header("Content-type", "application/json")
				self.end_headers()
				data = json.dumps(data).encode()
				self.wfile.write(data)

			elif path == "/api1/playlists":
				l = []
				for item in pctl.multi_playlist:
					p = {}
					p["name"] = item.title
					p["id"] = str(item.uuid_int)
					p["count"] = len(item.playlist_ids)
					l.append(p)
				data = {"playlists": l}
				self.send_response(200)
				self.send_header("Content-type", "application/json")
				self.end_headers()
				data = json.dumps(data).encode()
				self.wfile.write(data)

			elif path.startswith("/api1/albumtracks/"):
				# Get tracks that appear in an album /albumtracks/plid/albumid
				levels, _ = self.parse_trail(path)
				l = []
				if len(levels) == 5 and levels[3].isdigit() and levels[4].isdigit():
					pl = tauon.id_to_pl(int(levels[3]))
					if pl is not None:
						_, album, _ = tauon.get_album_info(int(levels[4]), pl)
						# logging.info(album)
						for p in album:
							l.append(self.get_track(p, pl, album_id=int(levels[4])))

				data = {"tracks": l}
				self.send_response(200)
				self.send_header("Content-type", "application/json")
				self.end_headers()
				data = json.dumps(data).encode()
				self.wfile.write(data)

			elif path.startswith("/api1/trackposition/"):
				# get track /trackposition/plid/playlistposition
				levels, _ = self.parse_trail(path)
				if len(levels) == 5 and levels[3].isdigit() and levels[4].isdigit():
					pl = tauon.id_to_pl(int(levels[3]))
					if pl is not None:
						data = self.get_track(int(levels[4]), pl)

						playlist = pctl.multi_playlist[pl].playlist_ids
						p = int(levels[4])
						if p < len(playlist):
							track = pctl.get_track(playlist[p])
							while True:
								if p < 0 or pctl.get_track(playlist[p]).parent_folder_path != track.parent_folder_path:
									p += 1
									break
								p -= 1
							data["album_id"] = p

							self.send_response(200)
							self.send_header("Content-type", "application/json")
							self.end_headers()
							data = json.dumps(data).encode()
							self.wfile.write(data)
						else:
							self.send_response(404)
							self.send_header("Content-type", "text/plain")
							self.end_headers()
							self.wfile.write(b"404 invalid track position")
					else:
						self.send_response(404)
						self.send_header("Content-type", "text/plain")
						self.end_headers()
						self.wfile.write(b"404 playlist not found")
				else:
					self.send_response(404)
					self.send_header("Content-type", "text/plain")
					self.end_headers()
					self.wfile.write(b"404 invalid track")

			elif path.startswith("/api1/setvolume/"):
				key = path[16:]
				if key.isdigit():
					volume = int(key)
					volume = max(volume, 0)
					volume = min(volume, 100)
					pctl.player_volume = volume
					pctl.set_volume()

				self.send_response(200)
				self.send_header("Content-type", "text/plain")
				self.end_headers()
				self.wfile.write(b"OK")

			elif path.startswith("/api1/setvolumerel/"):
				key = path[19:]
				if key.lstrip("-").isdigit() and key.count("-") < 2:
					volume = pctl.player_volume + int(key)
					volume = max(volume, 0)
					volume = min(volume, 100)
					pctl.player_volume = volume
					pctl.set_volume()

				self.send_response(200)
				self.send_header("Content-type", "text/plain")
				self.end_headers()
				self.wfile.write(b"OK")

			elif path.startswith("/api1/seek1k/"):
				key = path[13:]
				if key.isdigit():
					pctl.seek_decimal(int(key) / 1000)

				self.send_response(200)
				self.send_header("Content-type", "text/plain")
				self.end_headers()
				self.wfile.write(b"OK")

			elif path.startswith("/api1/seek/"):
				key = path[11:]
				if key.isdigit():
					pctl.seek_time(int(key) / 1000)

				self.send_response(200)
				self.send_header("Content-type", "text/plain")
				self.end_headers()
				self.wfile.write(b"OK")

			elif path.startswith("/api1/tracklist/"):
				# Return all tracks in a playlist /tracklist/plid
				key = path[16:]
				l = []
				if key.isdigit():
					pl = tauon.id_to_pl(int(key))
					if pl is not None and pl < len(pctl.multi_playlist):
						playlist = pctl.multi_playlist[pl].playlist_ids
						parent = ""
						album_id = 0
						for i, id in enumerate(playlist):
							tr = pctl.get_track(id)
							if i == 0:
								parent = tr.parent_folder_path
							elif parent != tr.parent_folder_path:
								parent = tr.parent_folder_path
								album_id = i

							l.append(self.get_track(i, pl, album_id=album_id))

				data = {"tracks": l}
				self.send_response(200)
				self.send_header("Content-type", "application/json")
				self.end_headers()
				data = json.dumps(data).encode()
				self.wfile.write(data)

			elif path.startswith("/api1/albums/"):
				# Returns lists of tracks that are start of albums /albums/plid
				key = path[13:]
				l = []
				if key.isdigit():
					pl = tauon.id_to_pl(int(key))
					if pl is not None:
						dex = tauon.reload_albums(True, pl)
						# logging.info(dex)
						for a in dex:
							l.append(self.get_track(a, pl, album_id=a))

				data = {"albums": l}
				self.send_response(200)
				self.send_header("Content-type", "application/json")
				self.end_headers()
				data = json.dumps(data).encode()
				self.wfile.write(data)

			elif path == "/api1/status":
				self.send_response(200)
				self.send_header("Content-type", "application/json")
				self.end_headers()
				data = {
					"status": "stopped",
					"inc": pctl.db_inc,
					"shuffle": pctl.random_mode is True,
					"album_shuffle": pctl.album_shuffle_mode is True,
					"repeat": pctl.repeat_mode is True,
					"progress": 0,
					"auto_stop": tauon.pctl.stop_mode != StopMode.OFF,
					"volume": pctl.player_volume,
					"playlist": str(tauon.get_playing_playlist_id()),
					"playlist_length": len(pctl.multi_playlist[pctl.active_playlist_playing].playlist_ids),
				}
				if pctl.playing_state == PlayingState.PLAYING:
					data["status"] = "playing"
				if pctl.playing_state == PlayingState.PAUSED:
					data["status"] = "paused"
				track = pctl.playing_object()
				if track:
					data["id"] = track.index
					data["title"] = track.title
					data["artist"] = track.artist
					data["album"] = track.album
					data["progress"] = round(pctl.playing_time * 1000)
					data["track"] = self.get_track(0, 0, track)

				p = pctl.playlist_playing_position
				data["position"] = p
				data["album_id"] = 0
				playlist = pctl.playing_playlist()

				if track is not None and p < len(playlist):
					while True:
						if p < 0 or pctl.get_track(playlist[p]).parent_folder_path != track.parent_folder_path:
							p += 1
							break
						p -= 1
					data["album_id"] = p

				data = json.dumps(data).encode()
				self.wfile.write(data)

			else:
				self.send_response(404)
				self.end_headers()
				self.wfile.write(b"404 Not found")
			tauon.wake()

	try:
		logging.info("Starting remote control server")
		httpd = ThreadedHTTPServer(("0.0.0.0", REMOTE_API_PORT), Server)
		discovery_publisher.start()
		httpd.serve_forever()
		httpd.server_close()
	except OSError as e:
		if str(e) == "[Errno 98] Address already in use":
			logging.error("Not starting web api server, is another Tauon instance already running?")  # noqa: TRY400
		else:
			logging.exception("Unknown OSError starting web api server!")
	except Exception:
		logging.exception("Failed starting web api server!")
	finally:
		discovery_publisher.stop()


def controller(tauon: Tauon) -> None:
	import base64

	class Server(BaseHTTPRequestHandler):
		def log_message(self, format, *args) -> None:
			logging.info(format % args)

		def do_GET(self) -> None:
			path = self.path
			if not path.startswith("/open/"):
				path = path.rstrip("/")

			if path == "/raise":
				tauon.request_raise()
			if path == "/reloadtheme":
				tauon.gui.reload_theme = True
				tauon.gui.update += 1
			if path == "/playpause":
				tauon.pctl.play_pause()
			if path == "/play":
				tauon.pctl.play()
			if path == "/pause":
				tauon.pctl.pause_only()
			if path == "/stop":
				tauon.pctl.stop()
			if path == "/next":
				tauon.pctl.advance()
			if path == "/previous":
				tauon.pctl.back()
			if path == "/shuffle":
				tauon.toggle_random()
			if path == "/repeat":
				tauon.toggle_repeat()
			if path.startswith("/open/"):
				rest = path[6:]
				try:
					path = base64.urlsafe_b64decode(rest.encode()).decode()
				except Exception:
					logging.exception("Invalid /open/ payload")
					self.send_response(400)
					self.end_headers()
					return
				tauon.open_uri(path)

			self.send_response(200)
			self.end_headers()
			tauon.wake()

	logging.info("Start controller server")
	try:
		httpd = HTTPServer(("127.0.0.1", 7813), Server)
		httpd.serve_forever()
		httpd.server_close()
	except OSError as e:
		if str(e) == "[Errno 98] Address already in use":
			logging.error("Not starting controller webserver, is another Tauon instance already running?")
		else:
			logging.exception("Unknown OSError starting controller server!")
	except Exception:
		logging.exception("Failed starting controller server!")


def authserve(tauon: Tauon) -> None:
	class Server(BaseHTTPRequestHandler):
		def log_message(self, format: str, *args) -> None:
			logging.info(format % args)

		def do_GET(self) -> None:
			path = self.path
			# if path.startswith("/tidalredir"):
			# 	self.send_response(200)
			# 	self.send_header("Content-type", "text/plain")
			# 	self.end_headers()
			#
			# 	tauon.tidal.login2(path)
			# 	self.wfile.write(b"You can close this now and return to Tauon Music Box")

			self.send_response(400)
			self.end_headers()

	httpd = HTTPServer(("127.0.0.1", 7811), Server)
	httpd.serve_forever()
	httpd.server_close()


def stream_proxy(tauon: Tauon) -> None:
	class Server(BaseHTTPRequestHandler):
		def log_message(self, format: str, *args: Any) -> None:
			logging.info(format % args)

		def do_GET(self) -> None:
			self.send_response(200)
			self.send_header("Content-type", "audio/ogg")
			self.end_headers()

			position = 0
			tauon.vb.reset()

			while True:
				if not tauon.stream_proxy.download_running:
					return

				while position < tauon.stream_proxy.c:
					if position not in tauon.stream_proxy.chunks:
						logging.info("The buffer was deleted too soon!")
						return

					self.wfile.write(tauon.stream_proxy.chunks[position])

					if tauon.prefs.backend == Backend.PHAZOR:
						tauon.vb.input(tauon.stream_proxy.chunks[position])

					position += 1

				time.sleep(0.01)

	httpd = HTTPServer(("127.0.0.1", 7812), Server)
	httpd.serve_forever()
	httpd.server_close()



"""Tauon Music Box - Tag Module

The purpose of this module is to read metadata from FLAC, OGG, OPUS, APE and WV files

There is additional parsing logic in in Tauon.tag_scan()
"""

# Copyright © 2015-2019, Taiko2k captain(dot)gxj(at)gmail.com

#     This file is part of Tauon Music Box.
#
#     Tauon Music Box is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     Tauon Music Box is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Lesser General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with Tauon Music Box.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

import io
import logging
import os
import re
import struct
import wave
from pathlib import Path
from typing import TYPE_CHECKING

from tauon.t_modules.t_extra import process_odat

if TYPE_CHECKING:
	from io import BufferedReader, BytesIO
	from types import TracebackType

	from typing_extensions import Self


LRC_TIMESTAMP_RE = re.compile(r"\[\d+:\d{1,2}\.\d{2,3}\]")
LRC_INFO_TAGS = (
	"[ti:",
	"[ar:",
	"[al:",
	"[au:",
	"[lr:",
	"[length:",
	"[by:",
	"[offset:",
	"[re:",
	"[tool:",
	"[ve:",
	"[#:",
)


def lyrics_are_synced(lyrics: str) -> bool:
	for line in lyrics.splitlines():
		line = line.strip().lower()
		if not line or line.startswith("#"):
			continue
		if any(line.startswith(tag) for tag in LRC_INFO_TAGS):
			continue
		if LRC_TIMESTAMP_RE.search(line):
			return True
	return False


def parse_mbids_from_vorbis(obj: Ape | Flac | Opus, key: str, value: str | bytes) -> bool:
	if key == "musicbrainz_artistid":
		if obj.musicbrainz_artistids is None:
			obj.musicbrainz_artistids = []
		obj.musicbrainz_artistids.append(value)
		return True

	if key == "musicbrainz_trackid":
		obj.musicbrainz_recordingid = value
		return True

	if key == "musicbrainz_releasetrackid":
		obj.musicbrainz_trackid = value
		return True

	if key == "musicbrainz_albumid":
		obj.musicbrainz_albumid = value
		return True

	if key == "musicbrainz_releasegroupid":
		obj.musicbrainz_releasegroupid = value
		return True

	return False


def parse_picture_block(f: BufferedReader) -> bytes:
	a = f.read(4)
	a = int.from_bytes(a, byteorder="big")
	# logging.info("Picture type: " + str(a))

	a = f.read(4)
	b = int.from_bytes(a, byteorder="big")
	# logging.info("MIME len: " + str(b))

	a = f.read(b)
	# logging.info(a)
	# logging.info("MIME: " + a.decode('ascii'))

	a = f.read(4)
	a = int.from_bytes(a, byteorder="big")
	# logging.info("Description len: " + str(a))

	a = f.read(a)
	# logging.info("Description: " + a.decode('utf-8'))

	a = f.read(4)
	# a = int.from_bytes(a, byteorder='big')
	# logging.info("Width: " + str(a))

	a = f.read(4)
	# a = int.from_bytes(a, byteorder='big')
	# logging.info("Height: " + str(a))

	a = f.read(4)
	# a = int.from_bytes(a, byteorder='big')
	# logging.info("BPP: " + str(a))

	a = f.read(4)
	# a = int.from_bytes(a, byteorder='big')
	# logging.info("Index colour: " + str(a))

	a = f.read(4)
	a = int.from_bytes(a, byteorder="big")
	# logging.info("Bin len: " + str(a))

	return f.read(a)


def parse_wavpack_header(header_data: bytes) -> tuple[int, float] | None:
	"""Parse a WavPack block header and return (sample_rate, length_seconds)."""
	if len(header_data) < 32:
		return None

	try:
		header = struct.unpack("<4cIH2B5I", header_data[:32])
	except struct.error:
		return None

	sample_rates = [
		6000,
		8000,
		9600,
		11025,
		12000,
		16000,
		22050,
		24000,
		32000,
		44100,
		48000,
		64000,
		88200,
		96000,
		192000,
	]  # Adapted from example in WavPack/cli/wvparser.c

	sample_rate_index = (header[11] & (15 << 23)) >> 23
	if sample_rate_index >= len(sample_rates):
		return None

	sample_rate = sample_rates[sample_rate_index]
	if sample_rate <= 0:
		return None

	length = header[8] / sample_rate
	return sample_rate, length


class TrackFile:
	"""Base class for codec classes"""

	def __init__(self) -> None:
		# fmt:off
		self.file: BufferedReader | None = None
		self.has_picture = False # Wav does not need this

		self.picture       = "" # Wav does not need this
		self.filepath      = ""
		self.album_artist  = ""
		self.artist        = ""
		self.genre         = ""
		self.date          = "" # Wav does not need this
		self.comment       = "" # Wav does not need this
		self.album         = ""
		self.track_number  = ""
		self.track_total   = "" # Wav does not need this
		self.title         = ""
		self.encoder       = "" # Wav does not need this
		self.disc_number   = "" # Wav does not need this
		self.disc_total    = "" # Wav does not need this
		self.lyrics        = "" # Wav does not need this
		self.synced_lyrics = ""
		self.composer      = "" # Wav does not need this

		# Extended metadata, named to mirror the matching TrackClass fields so
		# the importer can copy them straight across without a conversion table.
		# Wav does not need these.
		self.artists:                    list[str] | None = None
		self.album_artists:              list[str] | None = None
		self.artist_sort:                str | None = None
		self.genres:                     list[str] | None = None
		self.musicbrainz_artistids:      list[str] | None = None
		self.musicbrainz_recordingid:    str | None = None
		self.musicbrainz_trackid:        str | None = None
		self.musicbrainz_albumid:        str | None = None
		self.musicbrainz_releasegroupid: str | None = None
		self.replaygain_track_gain:      float | None = None
		self.replaygain_track_peak:      float | None = None
		self.replaygain_album_gain:      float | None = None
		self.replaygain_album_peak:      float | None = None
		self.FMPS_Rating:                float | None = None
		self.rdat:                       str | None = None

		self.sample_rate = 48000 # OPUS files are always 48000
		self.length = 0
		self.bit_rate = 0  # Wav does not need this
		self.bit_depth = 0 # Opus, M4a and Wav does not need this

		self.track_gain: float | None = None # Wav does not need this
		self.album_gain: float | None = None # Wav does not need this
		# fmt:on

	def __enter__(self) -> Self:
		"""Open the file when entering the context"""
		self.file = Path(self.filepath).open("rb")
		return self

	def __exit__(
		self,
		exc_type: type[BaseException] | None,
		exc_val: BaseException | None,
		exc_tb: TracebackType | None,
	) -> None:
		"""Close the file when exiting the context"""
		if self.file:
			self.file.close()
		self.file = None

	def set_vorbis_lyrics(self, key: str, value: bytes) -> None:
		lyrics = value.decode("utf-8")
		if key == "unsyncedlyrics":
			self.lyrics = lyrics
		elif key == "syncedlyrics":
			self.synced_lyrics = lyrics
		elif key == "lyrics":
			if lyrics_are_synced(lyrics):
				if not self.synced_lyrics:
					self.synced_lyrics = lyrics
			elif not self.lyrics:
				self.lyrics = lyrics


class Flac(TrackFile):
	def __init__(self, file: str) -> None:
		super().__init__()
		self.filepath = file
		self.cue_sheet = ""

	def read_vorbis(self, f: BufferedReader) -> None:
		block_position = 0

		buffer = f.read(4)
		block_position += 4
		jump = int.from_bytes(buffer, byteorder="little")

		f.read(jump)
		block_position += jump

		buffer = f.read(4)
		block_position += 4
		fields = int.from_bytes(buffer, byteorder="little")
		# logging.info(fields)
		album_artists: list[str] = []
		artists: list[str] = []
		genres: list[str] = []
		odat = ""

		for i in range(fields):
			buffer = f.read(4)
			block_position += 4

			jump = int.from_bytes(buffer, byteorder="little")
			buffer = f.read(jump)
			block_position += jump

			# logging.info(buffer.decode('utf-8'))

			position = 0

			while position < 40:
				# logging.info(sss[position:position+1])
				position += 1

				if buffer[position : position + 1] == b"=":
					a = buffer[0:position].decode("utf-8").lower()
					b = buffer[position + 1 :]

					if parse_mbids_from_vorbis(self, a, b.decode()):
						pass
						# logging.info("Found MBID data:")
						# logging.info(a)
						# logging.info(b)
					elif a == "genre":
						# self.genre = b.decode("utf-8")
						genres.append(b.decode())
					elif a == "cuesheet":
						self.cue_sheet = b.decode()
					elif a == "date":
						self.date = b.decode("utf-8")
					elif a == "originaldate":
						odat = b.decode("utf-8")
					elif a == "originalyear":
						if odat == "":
							odat = b.decode("utf-8")
					elif a == "comment":
						self.comment = b.decode("utf-8")
					elif a == "album":
						self.album = b.decode("utf-8")
					elif a == "title":
						self.title = b.decode("utf-8")
					elif a == "tracknumber":
						self.track_number = b.decode("utf-8")
					elif a in ("tracktotal", "totaltracks"):
						self.track_total = b.decode("utf-8")
					elif a == "encoder":
						self.encoder = b.decode("utf-8")
					elif a in ("albumartist", "album artist"):
						# self.album_artist = b.decode("utf-8")
						album_artists.append(b.decode("utf-8"))
					elif a == "artist":
						# self.artist = b.decode("utf-8")
						artists.append(b.decode())
					elif a in ("disctotal", "totaldiscs"):
						self.disc_total = b.decode("utf-8")
					elif a == "discnumber":
						self.disc_number = b.decode("utf-8")
					elif a == "metadata_block_picture":
						logging.info("Tag Scanner: Found picture inside vorbis comment inside a FLAC file. Ignoring")
						logging.info(f"      In file: {self.filepath}")
					elif a in ("lyrics", "syncedlyrics", "unsyncedlyrics"):
						self.set_vorbis_lyrics(a, b)
					elif a == "replaygain_track_gain":
						self.replaygain_track_gain = float(
							b.decode("utf-8").lower().strip(" db").replace(",", ".")
						)
					elif a == "replaygain_track_peak":
						self.replaygain_track_peak = float(b.decode("utf-8").replace(",", "."))
					elif a == "replaygain_album_gain":
						self.replaygain_album_gain = float(
							b.decode("utf-8").lower().strip(" db").replace(",", ".")
						)
					elif a == "replaygain_album_peak":
						self.replaygain_album_peak = float(b.decode("utf-8").replace(",", "."))
					elif a == "composer":
						self.composer = b.decode("utf-8")
					elif a == "fmps_rating":
						self.FMPS_Rating = float(b.decode("utf-8"))
					elif a == "artistsort":
						self.artist_sort = b.decode("utf-8")

					# else:
					# logging.info("Tag Scanner: Found unhandled FLAC Vorbis comment field: " + a)
					# logging.info(b)
					# logging.info("\n-------------------------------------------\n")

		f.seek(block_position * -1, 1)

		if album_artists:
			# self.album_artist = "; ".join(album_artists)
			self.album_artist = album_artists[0]
			if len(album_artists) > 1:
				self.album_artists = album_artists
		if artists:
			self.artist = "; ".join(artists)
			if len(artists) > 1:
				self.artists = artists
		if genres:
			self.genre = "; ".join(genres)
			if len(genres) > 1:
				self.genres = genres
		process_odat(self, odat)

	def read_seek_table(self, f: BufferedReader) -> None:
		f.read(10)
		buffer = f.read(8)
		a = int.from_bytes(buffer, byteorder="big")
		k = bin(a)[2:].zfill(64)

		self.sample_rate = int(k[0:20], 2)

		bps = int(k[23:28], 2)
		self.bit_depth = bps + 1

		samples = int(k[28:64], 2)

		self.length = samples / self.sample_rate
		f.seek(-18, 1)

	def read(self, get_picture: bool = False) -> None:
		if not self.file:
			self.file = Path(self.filepath).open("rb")
		f = self.file

		# Very helpful: https://xiph.org/flac/format.html
		size = os.path.getsize(self.filepath) / 8
		if size < 100:
			return

		s = f.read(4)

		# Find start of FLAC stream
		if s != b"fLaC":
			while f.tell() < size + 100:
				f.seek(-3, 1)
				s = f.read(4)
				if s == b"fLaC":
					break
			else:
				return

		i = 0
		while i < 20:
			i += 1
			z = self.read_block(f)

			if z[1] == 0:
				self.read_seek_table(f)
			if z[1] == 4:
				self.read_vorbis(f)
			if z[1] == 5:
				logging.info("Tag Scan: Flac file has native embedded CUE. Not supported")
				logging.info(f"      In file: {self.filepath}")
				# mark = f.tell()
				#
				# logging.info("Found flac cue")
				#
				# a = f.read(16*8)
				# a = int.from_bytes(a, byteorder='big')
				# logging.info(("Catalog Number: " + str(a)))
				#
				#
				# f.seek(mark)
				# f.seek(z[2], 1)

			if z[1] == 6 and get_picture:
				self.picture = parse_picture_block(f)
				self.has_picture = True
			else:
				f.read(z[2])

			if z[0] == 1:
				break

	def read_block(self, f: BufferedReader) -> tuple[int, int, int]:
		q = f.read(1)
		a = int.from_bytes(q, byteorder="big")
		k = bin(a)[2:].zfill(8)
		flag = int(k[:1], 2)
		block_type = int(k[1:], 2)
		s = f.read(3)
		a = int.from_bytes(s, byteorder="big")
		length = a

		return flag, block_type, length

	def get(self) -> None:
		pass


# file = 'b.flac'
#
# item = Flac(file)
# item.read()


class Opus(TrackFile):
	def __init__(self, file: str) -> None:
		super().__init__()
		self.filepath: str = file
		self.length = 0

	def get_more(self, f: BufferedReader, v: BytesIO) -> int:
		header = struct.unpack("<4sBBqIIiB", f.read(27))

		segs = struct.unpack("B" * header[7], f.read(header[7]))
		l = sum(segs)

		o = v.tell()
		v.seek(0, 2)
		v.write(f.read(l))
		v.seek(o)

		return l

	def read(self) -> None:
		if not self.file:
			self.file = Path(self.filepath).open("rb")
		f = self.file

		header = struct.unpack("<4sBBqIIiB", f.read(27))

		# logging.info(header)

		segs = struct.unpack("B" * header[7], f.read(header[7]))

		# l = sum(segs)
		# logging.info(f.read(l + 4))
		# f.seek(l * -1)

		s = f.read(7)

		if s == b"OpusHea":
			f.seek(-7, 1)
		elif s == b"\x01vorbis":
			s = f.read(4)
			a = struct.unpack("<B4i", f.read(17))  # 44100
			self.sample_rate = a[1]
			self.bit_rate = int(a[3] / 1000)
			f.seek(-28, 1)
		else:
			return

		for p in segs:
			f.read(p)

		v = io.BytesIO()
		v.seek(0)

		l = self.get_more(f, v)

		s = v.read(7)
		l -= 7

		if s == b"OpusTag":
			v.read(1)
			l -= 1
		elif s == b"\x03vorbis":
			pass
		else:
			return

		s = v.read(4)
		l -= 4
		a = int.from_bytes(s, byteorder="little")  # Vendor string length
		s = v.read(a)  # Vendor string
		l -= a

		s = v.read(4)
		l -= 4

		number = int.from_bytes(s, byteorder="little")  # Number of comments

		album_artists: list[str] = []
		artists: list[str] = []
		genres: list[str] = []
		odat = ""

		for i in range(number):
			s = v.read(4)
			l -= 4

			length = int.from_bytes(s, byteorder="little")

			while l < length + 4:
				l += self.get_more(f, v)

			s = v.read(length)
			l -= length

			position = 0
			while position < 40:
				position += 1

				if s[position : position + 1] == b"=":
					a = s[0:position].decode("utf-8").lower()
					b = s[position + 1 :]

					# logging.info(a)  # Key
					# logging.info(b)  # Value

					if parse_mbids_from_vorbis(self, a, b.decode()):
						pass
						# logging.info("Found MBID data:")
						# logging.info(a)
						# logging.info(b)

					elif a == "genre":
						# self.genre = b.decode("utf-8")
						genres.append(b.decode())
					elif a == "date":
						self.date = b.decode("utf-8")
					elif a == "originaldate":
						odat = b.decode("utf-8")
					elif a == "originalyear":
						if odat == "":
							odat = b.decode("utf-8")
					elif a == "comment":
						self.comment = b.decode("utf-8")
					elif a == "album":
						self.album = b.decode("utf-8")
					elif a == "title":
						self.title = b.decode("utf-8")
					elif a == "tracknumber":
						self.track_number = b.decode("utf-8")
					elif a in ("tracktotal", "totaltracks"):
						self.track_total = b.decode("utf-8")
					elif a == "encoder":
						self.encoder = b.decode("utf-8")
					elif a in ("albumartist", "album artist"):
						# self.album_artist = b.decode("utf-8")
						album_artists.append(b.decode("utf-8"))
					elif a == "artist":
						# self.artist = b.decode("utf-8")
						artists.append(b.decode())
					elif a == "metadata_block_picture":
						logging.info("Tag Scanner: Found picture in OGG/OPUS file.")
						logging.info(f"      In file: {self.filepath}")
						self.has_picture = True
						self.picture = b
						# logging.info(b)

					elif a == "replaygain_track_gain":
						self.replaygain_track_gain = float(b.decode("utf-8").lower().strip(" db"))
					elif a == "replaygain_track_peak":
						self.replaygain_track_peak = float(b.decode("utf-8"))
					elif a == "replaygain_album_gain":
						self.replaygain_album_gain = float(b.decode("utf-8").lower().strip(" db"))
					elif a == "replaygain_album_peak":
						self.replaygain_album_peak = float(b.decode("utf-8"))
					elif a == "discnumber":
						self.disc_number = b.decode("utf-8")
					elif a in ("disctotal", "totaldiscs"):
						self.disc_total = b.decode("utf-8")
					elif a in ("lyrics", "syncedlyrics", "unsyncedlyrics"):
						self.set_vorbis_lyrics(a, b)
					elif a == "composer":
						self.composer = b.decode("utf-8")
					elif a == "fmps_rating":
						self.FMPS_Rating = float(b.decode("utf-8"))
					elif a == "artistsort":
						self.artist_sort = b.decode("utf-8")
					# else:
					# logging.info("Tag Scanner: Found unhandled Vorbis comment field: " + a)
					# logging.info(b.decode("utf-8"))
					# logging.info("	  In file: " + self.filepath)

					break

		v.close()

		if album_artists:
			# self.album_artist = "; ".join(album_artists)
			self.album_artist = album_artists[0]
			if len(album_artists) > 1:
				self.album_artists = album_artists
		if artists:
			self.artist = "; ".join(artists)
			if len(artists) > 1:
				self.artists = artists
		if genres:
			self.genre = "; ".join(genres)
			if len(genres) > 1:
				self.genres = genres
		process_odat(self, odat)

		# Find the last Ogg page from end of file to get track length
		f.seek(-1, 2)

		# Crudely seek back for it
		g = 100000
		while g > 0:
			if f.tell() < 10:
				break
			g -= 1
			f.seek(-5, 1)
			s = f.read(4)

			if s == b"OggS":
				f.seek(-4, 1)
				header = struct.unpack("<4sBBqIIiB", f.read(27))
				self.length = header[3] / self.sample_rate
				break


# file = 'a.ogg'
#
# item = Opus(file)
# item.read()


class Ape(TrackFile):
	"""Helpful: http://wiki.hydrogenaud.io/index.php?title=APEv2_specification"""

	def __init__(self, file: str) -> None:
		super().__init__()
		self.filepath = file
		self.found_tag = False
		self.label = ""

	def read(self) -> None:
		if not self.file:
			self.file = Path(self.filepath).open("rb")
		a = self.file

		# Check size of file
		a.seek(0, 2)
		file_size = a.tell()

		# Get last 32 bytes where ape tag footer might be
		a.seek(-32, 1)
		b = a.read(32)
		footer = struct.unpack("<8c6i", b)

		# For use later
		found = 1

		# If its not an ape footer, seek through the file for a bit to see if we find it
		if b"".join(footer[0:8]) != b"APETAGEX":
			found = 0
			a.seek(0, 2)
			g = 1000
			while g > 0:
				if a.tell() < 1000:
					break
				g -= 1
				a.seek(-9, 1)
				s = a.read(8)

				# Found it
				if s == b"APETAGEX":
					found = 2
					a.seek(-8, 1)
					b = a.read(32)
					footer = struct.unpack("<8c6i", b)

		if found == 0:
			logging.info("Tag Scanner: Can't find APE tag")
		else:
			self.found_tag = True
			tag_len = footer[9]  # The size of the tag data (excludes header)
			num_items = footer[10]  # Number of fields in tag

			# logging.info("Tag len: " + str(tag_len))
			# logging.info("Items: " + str(num_items))

			# Seek to start of tag (after any header)
			if found == 1:
				a.seek(tag_len * -1, 2)
			else:
				a.seek(8 + (tag_len * -1), 1)

			# For every field in tag
			for q in range(num_items):
				# (field value length, 0=text 2=binary)
				ta = struct.unpack("<ii", a.read(8))

				# Collect every character until we reach null terminator
				name = b""
				for i in range(100):
					ch = a.read(1)
					if ch == b"\x00":
						break
					name += ch

				key = name.decode("utf-8").lower()
				# logging.info("Key: " + key)

				value = a.read(ta[0])
				# logging.info(value)

				if ta[1] == 0:
					value = value.decode("utf-8")
				elif ta[1] == 2:
					# Avoid decode of binary data
					pass

				# Fill in the class attributes
				if key == "title":
					self.title = value
				elif key == "artist":
					self.artist = value
				elif key == "genre":
					self.genre = value
				elif key == "discnumber":
					self.disc_number = value
				elif key == "disc":
					# Ape track fields appear to use fraction format, rather than separate fields for number and total
					# So we need to handle that here for consistency
					if "/" in value:
						self.disc_number, self.disc_total = value.split("/")
					else:
						self.disc_number = value

				elif key == "comment":
					self.comment = value
				elif key == "track":
					# Same deal as with disc number
					if "/" in value:
						self.track_number, self.track_total = value.split("/")
					else:
						self.track_number = value

				elif key == "year":
					self.date = value
				elif key == "album":
					self.album = value
				elif key == "artist":
					self.artist = value
				elif key == "composer":
					self.composer = value
				elif key in ("album artist", "albumartist"):
					self.album_artist = value
				elif key == "label":
					self.label = value
				elif key == "lyrics":
					self.lyrics = value
				elif key == "replaygain_track_gain":
					self.replaygain_track_gain = float(value.lower().strip(" db"))
				elif key == "replaygain_track_peak":
					self.replaygain_track_peak = float(value)
				elif key == "replaygain_album_gain":
					self.replaygain_album_gain = float(value.lower().strip(" db"))
				elif key == "replaygain_album_peak":
					self.replaygain_album_peak = float(value)
				elif parse_mbids_from_vorbis(self, key, value):
					pass
				elif key == "cover art (front)":
					# Data appears to have a filename at the start of it, we need to remove to recover a valid picture
					# Im not sure what the actual specification is here

					off = 0
					while off < 64:
						if value[off : off + 1] == b"\x00":
							off += 1
							break
						off += 1
					else:
						logging.info("Tag Scanner: Error reading APE album art")
						continue

					self.picture = value[off:]
					self.has_picture = True
					# logging.info(value)

		# Back to start of file to see if we can find sample rate and duration information
		a.seek(0)

		start = a.read(128)
		if start[0:3] == b"MAC":  # Ape files start with MAC
			version = struct.unpack("<h", start[4:6])[0]

			if version >= 3980:
				audio_info = struct.unpack("<IIIHHI", start[56:76])

				self.bit_depth = audio_info[3]
				self.sample_rate = audio_info[5]

				frames = audio_info[2] - 1
				blocks = audio_info[0]

				self.length = (frames * blocks) / self.sample_rate
			else:
				logging.info("Note: Old APE file format version")

		elif Path(self.filepath).suffix.lower() == ".tta":
			a.seek(0)
			signature = a.read(4)

			if signature == b"TTA1":
				header = a.read(18)
				if len(header) != 18:
					logging.info("WARNING: TTA1 header is incomplete")
				else:
					_audio_format, _channels, bps, sample_rate, total_samples, _crc32 = struct.unpack("<3H3L", header)
					if sample_rate > 0:
						self.sample_rate = sample_rate
					self.bit_depth = bps
					if self.sample_rate > 0:
						self.length = total_samples / self.sample_rate

			elif signature == b"TTA2":
				# TTA2 header:
				# channels(2), bits_per_sample(2), sample_rate(4), channel_mask(4),
				# total_samples(8), data_size(8)
				header = a.read(28)
				if len(header) != 28:
					logging.info("WARNING: TTA2 header is incomplete")
				else:
					_channels, bps, sample_rate, _channel_mask, total_samples, _data_size = struct.unpack(
						"<HHIIQQ", header
					)
					if sample_rate > 0:
						self.sample_rate = sample_rate
					self.bit_depth = bps
					if self.sample_rate > 0 and total_samples > 0:
						self.length = total_samples / self.sample_rate

			else:
				logging.info("WARNING: Does not appear to be a valid TTA file")
		elif Path(self.filepath).suffix.lower() == ".wv":
			#  We can handle WavPack files here too
			#  This code likely won't cover all cases as is, I only tested it on a few files

			a.seek(0)

			found_wvpk = False
			chunk_size = 64 * 1024
			scan_limit = max(file_size - 31, 0)  # Need 32 bytes remaining to parse a full header
			scan_pos = 0

			while scan_pos < scan_limit:
				a.seek(scan_pos)
				read_size = min(chunk_size, scan_limit - scan_pos)
				chunk = a.read(read_size)
				if not chunk:
					break

				search_from = 0
				while True:
					relative_pos = chunk.find(b"wvpk", search_from)
					if relative_pos == -1:
						break

					signature_pos = scan_pos + relative_pos
					a.seek(signature_pos)
					parsed = parse_wavpack_header(a.read(32))
					if parsed:
						self.sample_rate, self.length = parsed
						found_wvpk = True
						break
					search_from = relative_pos + 1

				if found_wvpk:
					break

				# Step by chunk minus 3 bytes so signatures split across chunk boundaries are still visible.
				if read_size <= 3:
					break
				scan_pos += read_size - 3

			if not found_wvpk:
				logging.info("Tag Scanner: Cannot verify WavPack file")
		else:
			logging.info("Tag Scanner: Does not appear to be an APE file")


class Wav(TrackFile):
	def __init__(self, file: str) -> None:
		super().__init__()
		self.filepath = file

	def read(self) -> None:
		if not self.file:
			self.file = Path(self.filepath).open("rb")
		f = self.file

		f.read(12)

		while True:
			wav_type = f.read(4)
			if not wav_type:
				break
			remain = int.from_bytes(f.read(4), "little")

			if wav_type != b"LIST":
				# RIFF chunks are word-aligned, so odd-sized chunks include a pad byte.
				f.seek(remain + (remain % 2), io.SEEK_CUR)
			else:
				info = f.read(4)
				if info == b"INFO":
					remain -= 4
					while remain > 0:
						tag_id = f.read(4).decode()
						size = int.from_bytes(f.read(4), "little")
						value = f.read(size)[:-1].decode("unicode_escape")
						if tag_id == "ITRK":
							self.track_number = value
						if tag_id == "IGNR":
							self.genre = value
						if tag_id == "IART":
							self.artist = value
						if tag_id == "INAM":
							self.title = value
						if tag_id == "IPRD":
							self.album = value

						if size % 2 == 1:
							size += 1
							f.read(1)

						remain -= 8 + size
				else:
					# Skip non-INFO LIST chunks (e.g. adtl) to avoid desyncing the chunk parser.
					remaining_list_bytes = max(remain - 4, 0)
					if remaining_list_bytes:
						f.seek(remaining_list_bytes, io.SEEK_CUR)
					if remain % 2 == 1:
						f.seek(1, io.SEEK_CUR)

		with wave.open(self.filepath, "rb") as wav:
			self.sample_rate = wav.getframerate()
			self.length = wav.getnframes() / self.sample_rate


genre_dict = {
	0: "Blues",
	1: "Classic Rock",
	2: "Country",
	3: "Dance",
	4: "Disco",
	5: "Funk",
	6: "Grunge",
	7: "Hip-Hop",
	8: "Jazz",
	9: "Metal",
	10: "New Age",
	11: "Oldies",
	12: "Other",
	13: "Pop",
	14: "R&B",
	15: "Rap",
	16: "Reggae",
	17: "Rock",
	18: "Techno",
	19: "Industrial",
	20: "Alternative",
	21: "Ska",
	22: "Death Metal",
	23: "Pranks",
	24: "Soundtrack",
	25: "Euro-Techno",
	26: "Ambient",
	27: "Trip-Hop",
	28: "Vocal",
	29: "Jazz+Funk",
	30: "Fusion",
	31: "Trance",
	32: "Classical",
	33: "Instrumental",
	34: "Acid",
	35: "House",
	36: "Game",
	37: "Sound Clip",
	38: "Gospel",
	39: "Noise",
	40: "Alternative Rock",
	41: "Bass",
	42: "Soul",
	43: "Punk",
	44: "Space",
	45: "Meditative",
	46: "Instrumental Pop",
	47: "Instrumental Rock",
	48: "Ethnic",
	49: "Gothic",
	50: "Darkwave",
	51: "Techno-Industrial",
	52: "Electronic",
	53: "Pop-Folk",
	54: "Eurodance",
	55: "Dream",
	56: "Southern Rock",
	57: "Comedy",
	58: "Cult",
	59: "Gangsta",
	60: "Top 40",
	61: "Christian Rap",
	62: "Pop/Funk",
	63: "Jungle",
	64: "Native US",
	65: "Cabaret",
	66: "New Wave",
	67: "Psychedelic",
	68: "Rave",
	69: "Showtunes",
	70: "Trailer",
	71: "Lo-Fi",
	72: "Tribal",
	73: "Acid Punk",
	74: "Acid Jazz",
	75: "Polka",
	76: "Retro",
	77: "Musical",
	78: "Rock & Roll",
	79: "Hard Rock",
	80: "Folk",
	81: "Folk-Rock",
	82: "National Folk",
	83: "Swing",
	84: "Fast Fusion",
	85: "Bebob",
	86: "Latin",
	87: "Revival",
	88: "Celtic",
	89: "Bluegrass",
	90: "Avantgarde",
	91: "Gothic Rock",
	92: "Progressive Rock",
	93: "Psychedelic Rock",
	94: "Symphonic Rock",
	95: "Slow Rock",
	96: "Big Band",
	97: "Chorus",
	98: "Easy Listening",
	99: "Acoustic",
	100: "Humour",
	101: "Speech",
	102: "Chanson",
	103: "Opera",
	104: "Chamber Music",
	105: "Sonata",
	106: "Symphony",
	107: "Booty Bass",
	108: "Primus",
	109: "Porn Groove",
	110: "Satire",
	111: "Slow Jam",
	112: "Club",
	113: "Tango",
	114: "Samba",
	115: "Folklore",
	116: "Ballad",
	117: "Power Ballad",
	118: "Rhythmic Soul",
	119: "Freestyle",
	120: "Duet",
	121: "Punk Rock",
	122: "Drum Solo",
	123: "Acapella",
	124: "Euro-House",
	125: "Dance Hall",
	126: "Goa",
	127: "Drum & Bass",
	128: "Club - House",
	129: "Hardcore",
	130: "Terror",
	131: "Indie",
	132: "BritPop",
	133: "Negerpunk",
	134: "Polsk Punk",
	135: "Beat",
	136: "Christian Gangsta Rap",
	137: "Heavy Metal",
	138: "Black Metal",
	139: "Crossover",
	140: "Contemporary Christian",
	141: "Christian Rock",
	142: "Merengue",
	143: "Salsa",
	144: "Thrash Metal",
	145: "Anime",
	146: "JPop",
	147: "Synthpop",
	148: "Unknown",
}


class M4a(TrackFile):
	def __init__(self, file: str) -> None:
		super().__init__()
		self.filepath = file
		self.sample_rate = 0  # Unknown sample rate

	def read(self, get_picture: bool = False) -> None:
		if not self.file:
			self.file = Path(self.filepath).open("rb")
		f = self.file

		k = [
			b"moov",
			b"trak",
			b"----",
			b"udta",
			b"meta",
			b"ilst",
			b"mdia",
			b"mdhd",
			b"minf",
			b"stbl",
			b"stsd",
			b"esds",
		]

		# s_name = b""

		def meta_get(f: BufferedReader, size: int) -> bytes:
			start = f.tell()
			f.seek(16, 1)
			data = f.read(size - 8 - 16)
			f.seek(start)
			return data

		def atom(f: BufferedReader, tail: bytes = b"", name: str | bytes = "") -> bool:
			# global s_name

			start = f.tell()
			b_size = f.read(4)
			size = int.from_bytes(b_size, "big")

			name = f.read(4)

			if name == b"":
				return False

			# logging.info(f"NAME: {tail}{b'.'}{name}")

			# Too lazy to parse each sub atom, lets just grab the data out the sub atom and
			# hope the file is formatted normally

			if name == b"\xa9nam":
				self.title = meta_get(f, size).decode().replace("\x00", "")

			if name == b"\xa9alb":
				self.album = meta_get(f, size).decode().replace("\x00", "")

			if name == b"\xa9ART":
				self.artist = meta_get(f, size).decode().replace("\x00", "")

			if name == b"\xa9wrt":
				self.composer = meta_get(f, size).decode().replace("\x00", "")

			if name == b"\xa9cmt":
				self.comment = meta_get(f, size).decode().replace("\x00", "")

			if name == b"\xa9lyr":
				self.lyrics = meta_get(f, size).decode().replace("\x00", "")

			if name == b"\xa9day":
				day = meta_get(f, size).decode().replace("\x00", "")
				if len(day) > 10 and day[10] == "T":
					day = day[:10]
				self.date = day

			if name == b"gnre":
				index = int.from_bytes(meta_get(f, size), "big")
				if index - 1 in genre_dict:
					self.genre = genre_dict[index - 1]

				# self.genre = meta_get(f, size).decode()

			if name == b"\xa9gen":
				self.genre = meta_get(f, size).decode().replace("\x00", "")

			if name == b"aART":
				self.album_artist = meta_get(f, size).decode().replace("\x00", "")

			if name == b"covr":
				self.has_picture = True
				if get_picture:
					self.picture = meta_get(f, size)

			if name == b"trkn":
				self.track_number = int(meta_get(f, size)[3])  # We'll just grab that

			if name == b"disk":  # They spelt disc wrong lol
				self.disc_number = int(meta_get(f, size)[3])

			# if tail[-4:] == b"----":
			# 	s_name: bytes = b""
			# 	if name == b"name":
			# 		s_name = f.read(size - 8)
			# 		f.seek((size - 8) * -1, 1)
			# 	elif name == b"data" and s_name != b"":
			# 		data = f.read(size - 8)
			# 		f.seek((size - 8) * -1, 1)
			# 		logging.info(s_name)
			# 		logging.info(data)

			if name in k:
				if name == b"----":
					f.seek(30, 1)

				if name == b"stsd":
					f.seek(44, 1)

				if name == b"meta":
					f.seek(4, 1)  # The 'meta' atom has some extra space at the start

				if name == b"mdhd":
					data = f.read(size - 8)
					f.seek((size - 8) * -1, 1)
					data = struct.unpack(">iiiiihh", data)
					self.sample_rate = data[3]
					self.length = data[4] / self.sample_rate

				if name == b"esds":
					f.seek(26, 1)
					bitrate = int.from_bytes(f.read(4), "big")
					f.seek(-30, 1)
					self.bit_rate = bitrate // 1000
				else:
					while f.tell() < start + size:
						if not atom(f, tail + b"." + name, name):
							break

			f.seek(start)
			f.seek(size, 1)
			return size != 0

		while atom(f):
			pass
